﻿'  This module is a work in progress - intended for adding a robotic arm.

'  VARIOUS DESIGN NOTES FROM WEB
'By default, it takes roughly 18ms to send or recieve a command to the motors using a USB2Dynamixel adapter
'In practise, communication is limited to roughly 70 times per second by default (ref: Mike Gebhard) even when using the full 1Mbps baud rate.

'According to a research study by Ethan Tira-Thompson (HERE), Dynamixel motors return the motor position within about half a degree of error 
'and at the full communication speed, but motor speed and "Present Load" torque readings only update every 130ms (just 7 times per second!) 
'and aren't nearly as accurate (roughly 10% error and have dead-zones). On the final page of the report, Ethan mentions that they have been 
'able to trick the controller into reading 3 sensor values within a single 16ms cycle (therefore running 20 motors at roughly 8 frames per 
'second instead of 3). 

'On the Robotis website, they posted a comment (http://www.robotis.com/zbxe/32886) saying that the main communication delay is caused by the 
' USB driver in Windows, which buffers data for 16ms by default. They say that you can adjust this down to 1ms, which would let you perform a 
'2-way ping on a Dynamixel motor at roughly 500Hz I think! I don't have any motors handy but it would be great if someone could try it to see 
'if this is actually true

'On Windows you use Device Manager->Ports->COM Port->Port Settings->Advanced->Latency Timer
Imports System.Math
Imports System.Threading

Module modMotionControl

    Dim timeStart As DateTime
    Dim boolMovementInProgress As Boolean = False
    Const constRampInterval As Single = 140
    Const constBaseMovementRatePerSecond As Single = 100
    Dim strDebug As String = ""
    Dim strDebug1 As String = ""

    'Public Sub subBaseRotate1(ByVal sngBaseAngleToMove As Single)
    '    On Error Resume Next
    '    Dim intBaseTimeToMove = sngBaseAngleToMove / 30 + constBaseMovementRatePerSecond
    '    Dim intRampInterval As Short = 2
    '    Dim sngCurrentMovementRate As Single = 0.1
    '    Dim boolStartRampDown As Boolean = False
    '    If (intRampInterval > Abs(intBaseTimeToMove) / 2) Then intRampInterval = Abs(intBaseTimeToMove) / 2
    '    Dim timeStart As DateTime = Now()
    '    strDebug = ""
    '    dxl_write_word(1, constMovingSpeed, sngCurrentMovementRate)
    '    ftnSetBaseRotatePosition(intBaseIntervalToMove)
    '    Do

    '        If (ftnElapsedTime_S(timeStart) 

    '            dxl_write_word(1, constMovingSpeed, sngCurrentMovementRate)
    '    Loop Until (sngCurrentMovementRate <= 0.05)
    '    MsgBox(strDebug)
    'End Sub

    Public Sub subBaseRotate(ByVal sngBaseAngleToMove As Single)
        On Error Resume Next
        Dim intBaseIntervalToMove = sngBaseAngleToMove * 1024 / 300
        Dim intBaseStartPosition As Short = dxl_read_word(1, constPresentPosition)
        Dim intBaseFinalPosition As Short = intBaseStartPosition + intBaseIntervalToMove
        Dim sngBaseGoalPosition As Single = intBaseStartPosition
        Dim intRampInterval As Short = constRampInterval
        Dim sngCurrentMovementRate As Single = 0.1
        Dim boolStartRampDown As Boolean = False
        Dim intRampShortened As Short = 0
        If (intRampInterval > Abs(intBaseIntervalToMove) / 2) Then
            intRampInterval = Abs(intBaseIntervalToMove) / 2
            intRampShortened = constRampInterval - intRampInterval
        End If
        Dim timeLoop As DateTime = Now()
        Dim intPositionIncrement As Short
        strDebug = ""
        Do
            strDebug1 = ftnElapsedTime_S(timeLoop) & " : "
            If (intBaseIntervalToMove > 0) And (sngBaseGoalPosition < (intBaseStartPosition + intRampInterval)) Then  'Ramp Up CCW
                sngCurrentMovementRate = (ftnRampValue(Abs(sngBaseGoalPosition - intBaseStartPosition), constRampInterval) * constBaseMovementRatePerSecond - sngCurrentMovementRate * 0) '* ftnElapsedTime_S(timeLoop)
                strDebug1 &= "----UP----"
            ElseIf (intBaseIntervalToMove < 0) And (sngBaseGoalPosition > (intBaseStartPosition - intRampInterval)) Then  'Ramp Up CW
                sngCurrentMovementRate = (ftnRampValue(Abs(sngBaseGoalPosition - intBaseStartPosition), constRampInterval) * constBaseMovementRatePerSecond - sngCurrentMovementRate * 0) '* ftnElapsedTime_S(timeLoop)
                strDebug1 &= "----UP----"
            ElseIf (intBaseIntervalToMove > 0) And (sngBaseGoalPosition > (intBaseFinalPosition - intRampInterval)) Then 'Ramp Down CCW
                If Not boolStartRampDown Then
                    boolStartRampDown = True
                    intRampShortened = intRampShortened * Abs(sngBaseGoalPosition - intBaseFinalPosition) / intRampInterval
                    intRampInterval = Abs(sngBaseGoalPosition - intBaseFinalPosition)
                    strDebug = " DN PREP:  Interval = " & intRampInterval & "  Shortened = " & intRampShortened & vbCrLf & strDebug
                End If
                sngCurrentMovementRate = (ftnRampValue(Abs(sngBaseGoalPosition - intBaseFinalPosition) + intRampShortened, constRampInterval) * constBaseMovementRatePerSecond - sngCurrentMovementRate * 0) ' * ftnElapsedTime_S(timeLoop)
                strDebug1 &= "----DN----"
            ElseIf (intBaseIntervalToMove < 0) And (sngBaseGoalPosition < (intBaseFinalPosition + intRampInterval)) Then 'Ramp Down CW
                If Not boolStartRampDown Then
                    boolStartRampDown = True
                    intRampShortened = intRampShortened * Abs(sngBaseGoalPosition - intBaseFinalPosition) / intRampInterval
                    intRampInterval = Abs(sngBaseGoalPosition - intBaseFinalPosition)
                    strDebug = "DN PREP:  Interval = " & intRampInterval & "  Shortened = " & intRampShortened & vbCrLf & strDebug
                End If
                sngCurrentMovementRate = (ftnRampValue(Abs(sngBaseGoalPosition - intBaseFinalPosition) + intRampShortened, constRampInterval) * constBaseMovementRatePerSecond - sngCurrentMovementRate * 0) '* ftnElapsedTime_S(timeLoop)
                strDebug1 &= "----DN----"
            Else
                strDebug1 &= "----FL----"
            End If
            intPositionIncrement = sngCurrentMovementRate * ftnElapsedTime_S(timeLoop)
            If (intPositionIncrement = 0) Then intPositionIncrement = 1
            If (intBaseIntervalToMove > 0) Then
                sngBaseGoalPosition += intPositionIncrement
                strDebug1 &= " Rate = " & sngCurrentMovementRate & "   "
                If (sngBaseGoalPosition > intBaseFinalPosition) Then sngBaseGoalPosition = intBaseFinalPosition
            Else
                sngBaseGoalPosition -= intPositionIncrement
                strDebug1 &= "  Rate = " & sngCurrentMovementRate & "   "
                If (sngBaseGoalPosition < intBaseFinalPosition) Then sngBaseGoalPosition = intBaseFinalPosition
            End If
            timeLoop = Now()
            strDebug1 &= ", Goal = " & (sngBaseGoalPosition * 300 / 1024 - 150).ToString & vbCrLf
            strDebug = strDebug1 & strDebug
            ftnSetBaseRotatePosition(sngBaseGoalPosition)
        Loop Until (sngBaseGoalPosition = intBaseFinalPosition)
        'MsgBox(strDebug)
    End Sub




    Function ftnRampValue(ByVal sngCurrentPosition As Single, ByVal sngRampInterval As Single) As Single
        On Error Resume Next
        Dim sngRampFactor As Single = sngCurrentPosition / sngRampInterval
        If (sngRampFactor > 1) Then sngRampFactor = 1
        ftnRampValue = (3 - 2 * sngRampFactor) * sngRampFactor * sngRampFactor
        strDebug1 &= " Ramp Position = " & sngCurrentPosition & " Interval = " & sngRampInterval & " Ramp = " & ftnRampValue
    End Function

    Public Function ftnElapsedTime_S(ByVal timeStart As DateTime) As Single ' returns elapsed time since timeStart in seconds
        On Error Resume Next
        Dim tsElapsedTimeSpan As TimeSpan = Now().Subtract(timeStart)
        ftnElapsedTime_S = tsElapsedTimeSpan.TotalMilliseconds / 1000
        If (ftnElapsedTime_S < 0) Then ftnElapsedTime_S = 0
    End Function

    Public Sub subWait(ByVal intMillseconds As Integer)
        On Error Resume Next
        Dim TimeEnd As DateTime
        TimeEnd = Now().AddMilliseconds(intMillseconds)
        Do
            System.Windows.Forms.Application.DoEvents()
        Loop Until Now() > TimeEnd
        TimeEnd = Nothing
    End Sub

End Module
