﻿
' Module which controls the driving of the robot. 

Imports System.Math
Imports System.Threading


Module modBotMovementControl

    Dim tbpLastLocalizedPose As typeBotPose
    Dim intConsecutiveTurnCounter As Short = 0  ' Counts the consecutive number of "Turn" moves.
    '                                                  Used to prevent excessive localization and new path generation 


    Public Sub subInitializeDrivingControl(ByRef tbpEstimatedPose As typeBotPose, ByRef tbpStartingPose As typeBotPose)
        If Not boolDrivingToDestination Then 'initialize
            intCollisions = 0
            boolBotIsMoving = False
            boolPoseConfirmedAfterCollision = True ' 
            tbpLastLocalizedPose = tbpEstimatedPose
            intMovementsSinceLocalization = 0
            ftnLocalizeUsingOdometry(tbpEstimatedPose, tbpStartingPose) ' initialize tmcStartingOdometry
            intConsecutiveTurnCounter = 0
        End If
    End Sub

    Public Function ftnBotStop(ByRef tbpEstimatedPose As typeBotPose, ByRef tbpTargetPose As typeBotPose) As typeBotPose
        Try
            strStatusDriving = intDrivingSequenceNumber & " : Stop: "
            ftnBotStop = tbpEstimatedPose
            ftnExecuteMove(tbpEstimatedPose, "stop", tbpTargetPose)
            GotABotClass.subRedraw(ftnBotStop)
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnBotStop")
        End Try
    End Function


    Function ftnBotFwd(ByRef tbpEstimatedPose As typeBotPose, ByVal tbpTargetPose As typeBotPose) As typeBotPose
        Try
            ftnBotFwd = tbpEstimatedPose
            If boolDebugDrive Then
                frmDebug.subWriteLogs(Color.Black, "        : ftnBotFwd   tbpEstimatedPose =: " & tbpEstimatedPose.Location.X & ", " & tbpEstimatedPose.Location.Y & "  at " & tbpEstimatedPose.Orientation)
                frmDebug.subWriteLogs(Color.Black, "        : ftnBotFwd   tbpTargetPose =: " & tbpTargetPose.Location.X & ", " & tbpTargetPose.Location.Y & "  at " & tbpTargetPose.Orientation)
            End If
            If Not boolStop Then
                strTypeOfMove = "Fwd"
                Dim sngDistanceToTarget As Single = ftnDistanceCommanded(tbpEstimatedPose, tbpTargetPose)
                If sngDistanceToTarget = 0 Then
                    sngDistanceToTarget = sngTwoInches 'Error Recovery
                    tbpTargetPose.Location = ftnLocationFromEstimatedPose(tbpEstimatedPose, 0, sngDistanceToTarget)
                    frmDebug.subWriteLogs(Color.Red, ftnElapsedTime(timeMoveCommandSent) & " ***** Path Failure:  Fwd 2 commanded ***************************")
                ElseIf (sngDistanceToTarget > sngMaxTravel) Then
                    sngDistanceToTarget = sngMaxTravel
                    tbpTargetPose.Location = ftnLocationFromEstimatedPose(tbpEstimatedPose, 0, sngDistanceToTarget)
                End If
#If constCompile2dAccelerometer Then
ftnTransmitGotAKinect("get accelerometer distance ", "OK", 1500) ' start GotAKinects distance accumulation
#End If
                intConsecutiveTurnCounter = 0
                If strUnits = "inch" Then
                    ftnBotFwd = ftnExecuteMove(tbpEstimatedPose, "move " & sngDistanceToTarget & " i", tbpTargetPose)
                Else
                    ftnBotFwd = ftnExecuteMove(tbpEstimatedPose, "move " & sngDistanceToTarget & " cm", tbpTargetPose)
                End If
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnBotFwd")
        End Try
    End Function

    Function ftnBotXY(ByRef tbpEstimatedPose As typeBotPose, ByVal tbpTargetPose As typeBotPose) As typeBotPose
        Try
            ftnBotXY = tbpEstimatedPose
            If boolDebugDrive Then
                frmDebug.subWriteLogs(Color.Black, "        : ftnBotXY   tbpEstimatedPose =: " & tbpEstimatedPose.Location.X & ", " & tbpEstimatedPose.Location.Y & "  at " & tbpEstimatedPose.Orientation)
                frmDebug.subWriteLogs(Color.Black, "        : ftnBotXY   tbpTargetPose =: " & tbpTargetPose.Location.X & ", " & tbpTargetPose.Location.Y & "  at " & tbpTargetPose.Orientation)
            End If
            If Not boolStop Then
                strTypeOfMove = "XY"
                Dim tmcXY_Distances As typeMapCoordinate = ftnFindXY_Parameters(tbpEstimatedPose, tbpTargetPose)
                Dim sngDistanceToTarget As Single = ftnDistanceBetweenLocations(tbpEstimatedPose.Location, tbpTargetPose.Location)
                If sngDistanceToTarget = 0 Then
                    frmDebug.subWriteLogs(Color.Red, ftnElapsedTime(timeMoveCommandSent) & " ***** Path Failure - move XY:  move  2  i commanded ***************************")
                    tbpTargetPose.Location = ftnLocationFromEstimatedPose(tbpEstimatedPose, 0, sngTwoInches)
                    ftnBotXY = ftnBotFwd(tbpEstimatedPose, tbpTargetPose)
                ElseIf (sngDistanceToTarget > sngMaxTravel * 1.5) Or Not boolEnableXY_Motion Then
                    If ftnTargetRequiresTurn(tbpEstimatedPose, tbpTargetPose.Location) Then
                        ftnBotXY = ftnBotTurn(tbpEstimatedPose, True, tbpTargetPose)
                    Else
                        ftnBotXY = ftnBotFwd(tbpEstimatedPose, tbpTargetPose)
                    End If
                ElseIf (Int(tmcXY_Distances.Y) = 0) Then
                    ftnBotXY = ftnBotFwd(tbpEstimatedPose, tbpTargetPose)
                Else
                    intConsecutiveTurnCounter = 0
                    If strUnits = "inch" Then
                        ftnBotXY = ftnExecuteMove(tbpEstimatedPose, "move X " & Int(tmcXY_Distances.X) & " Y " & Int(tmcXY_Distances.Y) & " i", tbpTargetPose)
                    Else
                        ftnBotXY = ftnExecuteMove(tbpEstimatedPose, "move X " & Int(tmcXY_Distances.X) & " Y " & Int(tmcXY_Distances.Y) & " cm", tbpTargetPose)
                    End If
                End If
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnBotXY")
        End Try
    End Function

    Function ftnBotTurn(ByRef tbpEstimatedPose As typeBotPose, ByRef boolRestrictConsecutiveTurns As Boolean, _
                         ByVal tbpTargetPose As typeBotPose) As typeBotPose
        Try
            ftnBotTurn = tbpEstimatedPose
            If boolDebugDrive Then
                frmDebug.subWriteLogs(Color.Black, "        : ftnBotTurn   tbpEstimatedPose =: " & tbpEstimatedPose.Location.X & ", " & tbpEstimatedPose.Location.Y & "  at " & tbpEstimatedPose.Orientation)
                frmDebug.subWriteLogs(Color.Black, "        : ftnBotTurn   tbpTargetPose =: " & tbpTargetPose.Location.X & ", " & tbpTargetPose.Location.Y & "  at " & tbpTargetPose.Orientation)
            End If
            If Not boolStop Then
                strTypeOfMove = "Turn"
                If boolRestrictConsecutiveTurns And (intConsecutiveTurnCounter >= constMaxTurnsWithLocalization) Then ' intConsecutiveTurnCounter keeps track of turns and small movements and is used to limit unproductive localizations and path regeneration
                    frmDebug.subWriteLogs(Color.Red, ftnElapsedTime(timeMoveCommandSent) & " *********** Too many consecutive Turns - Moving forward instead ***************************")
                    tbpTargetPose = tbpEstimatedPose
                    tbpTargetPose.Location = ftnLocationFromEstimatedPose(tbpEstimatedPose, 0, sngTwoInches)
                    ftnBotTurn = ftnBotFwd(tbpEstimatedPose, tbpTargetPose)
                Else
                    If Not ftnTargetRequiresTurn(tbpEstimatedPose, tbpTargetPose.Location) Then
                        frmDebug.subWriteLogs(Color.Red, ftnElapsedTime(timeMoveCommandSent) & " *********** Path Failure Turn - Moving forward instead ***************************")
                        If boolDebugPath Then
                            frmDebug.subWriteLogs(Color.Black, "  :    tbpEstimatedPose =: " & tbpEstimatedPose.Location.X & ", " & tbpEstimatedPose.Location.Y & "  at " & tbpEstimatedPose.Orientation)
                            frmDebug.subWriteLogs(Color.Black, "  :    tbpTargetPose =: " & tbpTargetPose.Location.X & ", " & tbpTargetPose.Location.Y & "  at " & tbpTargetPose.Orientation)
                            frmDebug.subWriteLogs(Color.Black, "  :    tmcDestination =: " & tmcDestination.X & ", " & tmcDestination.Y)
                        End If
                        Dim tbpFindTargetPose As typeBotPose = ftnFindTargetPose(tbpEstimatedPose, tmcDestination)
                        frmDebug.subWriteLogs(Color.Black, "  :    tbpFindTargetPose =: " & tbpFindTargetPose.Location.X & ", " & tbpFindTargetPose.Location.Y & "  at " & tbpFindTargetPose.Orientation)
                        tbpTargetPose = tbpEstimatedPose
                        tbpTargetPose.Location = ftnLocationFromEstimatedPose(tbpEstimatedPose, 0, sngTwoInches)
                        ftnBotTurn = ftnBotFwd(tbpEstimatedPose, tbpTargetPose)
                    Else
                        Dim intTurnAngle As Short = ftnTurnAngle(tbpEstimatedPose, tbpTargetPose)
                        If (Abs(intTurnAngle) > sngMaxTurn) Then
                            If (intTurnAngle > 0) Then
                                intTurnAngle = sngMaxTurn
                            Else
                                intTurnAngle = -sngMaxTurn
                            End If
                        End If
                        If (Abs(intTurnAngle) > 135) Then intTurnAngle = ftnTurnWithoutHittingTail(tbpEstimatedPose, intTurnAngle)
                        tbpTargetPose.Orientation = ftnConvertTo360(tbpEstimatedPose.Orientation + intTurnAngle)
                        ftnBotTurn = ftnExecuteMove(tbpEstimatedPose, "move " & intTurnAngle & " degrees", tbpTargetPose)
                        intConsecutiveTurnCounter += 1
                    End If
                End If
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnBotTurn")
        End Try
    End Function

    Function ftnTurnWithoutHittingTail(ByRef tbpEstimatedPose As typeBotPose, ByRef intTurnAngle As Short) As Single
        ' This routine assumes that the robot is not perfectly round but rather that the robot's tail extends
        ' furthur from the center of movement than the head or sides.  To lessen the chances of collision between
        ' the tail and a nearby wall while turning, this routine will reverse the turn direction when near a wall 
        ' and when the direction of turn moves the tail towards the wall.
        Try
            ftnTurnWithoutHittingTail = intTurnAngle
            If Not boolSlam Then
                Dim boolNearWallLeft As Boolean = False
                Dim boolNearWallRight As Boolean = False
                Dim boolNearWallRear As Boolean = False
                Dim tmcCheckLocation As typeMapCoordinate
                tmcCheckLocation = ftnLocationFromLocation(tmcCheckLocation, ftnConvertTo360(tbpEstimatedPose.Orientation - 180), sngBotRadius)
                boolNearWallRear = boolOccupiedCell(tmcCheckLocation.X, tmcCheckLocation.Y)
                If Not boolNearWallRear Then
                    Dim sngDistance As Single = sngBotRadius
                    While Not boolNearWallLeft And Not boolNearWallRight And (sngDistance < (sngBotRadius + sngSixInches)) ' Find if there is a wall near the sides of the bot
                        tmcCheckLocation = ftnLocationFromEstimatedPose(tbpEstimatedPose, 90, sngDistance)
                        boolNearWallLeft = boolOccupiedCell(tmcCheckLocation.X, tmcCheckLocation.Y)
                        tmcCheckLocation = ftnLocationFromEstimatedPose(tbpEstimatedPose, -90, sngDistance)
                        boolNearWallRight = boolOccupiedCell(tmcCheckLocation.X, tmcCheckLocation.Y)
                        sngDistance += sngTwoInches
                    End While
                    If boolNearWallLeft And Not boolNearWallRight And (intTurnAngle < 0) Then
                        ftnTurnWithoutHittingTail = 360 + intTurnAngle
                    ElseIf boolNearWallRight And Not boolNearWallLeft And (intTurnAngle > 0) Then
                        ftnTurnWithoutHittingTail = -(360 - intTurnAngle)
                    Else
                        ftnTurnWithoutHittingTail = intTurnAngle
                    End If
                End If
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnTurnWithoutHittingTail")
        End Try
    End Function

    Dim boolMoveComplete As Boolean = False
    Dim boolCourseCorrectInProgress As Boolean = False
    Dim boolCourseCorrectCnt As Short = 0
    Function ftnExecuteMove(ByRef tbpEstimatedPose As typeBotPose, ByRef strCommand As String, ByVal tbpTargetPose As typeBotPose) As typeBotPose ' Execute move and update pose using dead reckoning and odometry
        Try
            ftnExecuteMove = tbpEstimatedPose
            If boolDebugDrive Then frmDebug.subWriteLogs(Color.Green, ftnElapsedTime(timeMoveCommandSent) & " ftnExecuteMove Start. strCommand =   " & strCommand)
            If (GotABotClass.picStop.BorderStyle = FormBorderStyle.Fixed3D) Then boolStop = True
            Dim boolCourseCorrectionMove As Boolean = False
            boolRangeReadingsAreFresh = False
            If boolSimulatedMovement Then
                ftnExecuteMove = ftnSimulatedExecuteMove(tbpEstimatedPose, strCommand, tbpTargetPose)
            ElseIf Not boolShutDown Then
                If (Not boolMoveComplete Or boolBotIsMoving) And (boolStop Or (String.CompareOrdinal(strCommand, "stop") = 0)) Or (GotABotClass.picStop.BorderStyle = FormBorderStyle.Fixed3D) Then   ' Stop movement
                    boolCourseCorrectInProgress = True
                    boolCourseCorrectCnt += 1
                    ftnExecuteMove = ftnExecuteStopCommand(tbpEstimatedPose, tbpTargetPose)
                    tbpTargetPoseGlobal = ftnExecuteMove
                ElseIf (String.CompareOrdinal(strTypeOfMove, "Fwd") = 0) Or (String.CompareOrdinal(strTypeOfMove, "XY") = 0) Or (String.CompareOrdinal(strTypeOfMove, "Turn") = 0) And Not boolStop Then
                    If Not (boolSlam And boolLocalizationInProgress) Then
                        tbpTargetPoseGlobal = tbpTargetPose
                        If Not boolBotIsMoving Then  ' New move / not a course correction
                            boolCourseCorrectInProgress = False
                            boolCollisionPredicted = False
                            boolCourseCorrectCnt = 0
                            frmDebug.subWriteMoveStartLogFile(strCommand, tbpTargetPose)
                            ftnExecuteMove = ftnExecuteMoveCommand(tbpEstimatedPose, tbpTargetPose, strCommand, tbpStartingPose)
                        Else
                            boolCourseCorrectionMove = True
                            boolCourseCorrectCnt += 1
                            If Not boolMoveComplete Then  ' course correction
                                boolCourseCorrectInProgress = True
                                ftnExecuteMove = ftnExecuteMoveCommand(tbpEstimatedPose, tbpTargetPose, strCommand, tbpStartingPose)
                            End If
                        End If
                        ftnExecuteMove = ftnExecuteMoveComplete(ftnExecuteMove, boolCourseCorrectionMove)
                    End If
                End If
            End If
            If boolDebugDrive Then frmDebug.subWriteLogs(Color.Green, ftnElapsedTime(timeMoveCommandSent) & " ftnExecuteMove End. strCommand =   " & strCommand)
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnExecuteMove")
        End Try
    End Function

    Function ftnExecuteMoveComplete(ByRef tbpEstimatedPose As typeBotPose, ByRef boolCourseCorrectionMove As Boolean) As typeBotPose ' Complete ftnExecuteMove with Dead Reckoning and Odometry localization
        Try
            If Not boolMoveComplete And Not boolCourseCorrectionMove Then
                Dim timeOut As DateTime = Now().AddMilliseconds(sngMaxTravel / ftnCmToMapUnits(sngBotLinearVelocity))
                While boolCourseCorrectInProgress And (Now() < timeOut) And (boolCourseCorrectCnt <> 0)  ' Wait for Course correction to complete
                    Thread.Sleep(25)
                End While
                boolMoveComplete = True
                ftnExecuteMoveComplete = ftnLocalizeUsingDeadReckoning(tbpEstimatedPose, 0, timeMoveCommandSent, tbpStartingPose, tbpTargetPoseGlobal, strTypeOfMove, True)
                frmDebug.subWritePose(ftnElapsedTime(timeMoveCommandSent) & "       Pose Derived From Dead Reckoning", ftnExecuteMoveComplete)
                ftnExecuteMoveComplete = ftnLocalizeUsingOdometry(ftnExecuteMoveComplete, tbpStartingPose)
                tbpTargetPoseGlobal = ftnExecuteMoveComplete
                If boolEnableLocalizationWhileMoving Then subEndImageRecognition() '
                subGenerateProposalFromMotionModel(ftnExecuteMoveComplete, tbpStartingPose, strTypeOfMove) '
                tbpStartingPose = tbpTargetPoseGlobal
                boolRangeReadingsAreFresh = False
                If boolDebugDrive Then frmDebug.subWriteLogs(Color.Green, ftnElapsedTime(timeMoveCommandSent) & " ftnExecuteMove Move complete:  ")
                intDrivingSequenceNumber += 1
                boolCourseCorrectCnt = 0
                boolBotIsMoving = False
                boolCourseCorrectInProgress = False
            Else
                ftnLocalizeUsingOdometry(ftnExecuteMoveComplete, tbpStartingPose)
                tbpStartingPose = tbpTargetPoseGlobal
                If (boolCourseCorrectCnt = 0) Then boolCourseCorrectInProgress = False
                boolCourseCorrectCnt -= 1
                Dim timeOut As DateTime = Now().AddMilliseconds(sngMaxTravel / ftnCmToMapUnits(sngBotLinearVelocity))
                While boolBotIsMoving And (Now() < timeOut)  ' wait for original move command to complete
                    Thread.Sleep(25)
                End While
                boolCourseCorrectCnt = 0
                ftnExecuteMoveComplete = tbpTargetPoseGlobal
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in subExecuteMoveComplete")
        End Try
    End Function

    Function ftnExecuteStopCommand(ByRef tbpEstimatedPose As typeBotPose, ByRef tbpTargetPose As typeBotPose) As typeBotPose
        Try
            boolCourseCorrectInProgress = True
            If boolDebugDrive Then frmDebug.subWriteLogs(Color.Green, ftnElapsedTime(timeMoveCommandSent) & "  ftnExecuteStopCommand Start:  ")
            ftnExecuteStopCommand = tbpEstimatedPose
            Dim strResponse As String = ftnSendCommand("stop")
            frmDebug.subWriteLogs(Color.Red, ftnElapsedTime(timeMoveCommandSent) & " ************************ STOP commanded  ********************************")
            If boolDebugDrive Then frmDebug.subWritePose(ftnElapsedTime(timeMoveCommandSent) & "       tbpEstimatedPose", tbpEstimatedPose)
            If boolBotIsMoving Then
                ftnExecuteStopCommand = ftnCheckForMoveComError(tbpEstimatedPose, strResponse, tbpTargetPose)
                ftnExecuteStopCommand.LocationUncertainty.X += sngFourInchesGrid
                ftnExecuteStopCommand.LocationUncertainty.Y += sngFourInchesGrid
                ftnExecuteStopCommand.OrientationUncertainty += 8
                intMovementsSinceLocalization = 99  ' force localization 
            End If
            If boolCollisionDetected Then
                subDrawScanVectors(ftnExecuteStopCommand, False)
                frmDebug.subWriteLogs(Color.Red, "                (Collision was detected)")
                ftnSendCommand("play phraseA 'Emergency Stop.  '")
            ElseIf boolCollisionPredicted Then
                subDrawScanVectors(ftnExecuteStopCommand, False)
                frmDebug.subWriteLogs(Color.Red, "                (Collision was predicted)")
                ftnSendCommand("play phraseA 'Emergency Stop.  '")
            End If
            boolCollisionPredicted = False
            If boolDebugDrive Then frmDebug.subWritePose(ftnElapsedTime(timeMoveCommandSent) & "       ftnExecuteStopCommand", ftnExecuteStopCommand)
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnExecuteStopCommand")
        End Try
    End Function

    Function ftnExecuteMoveCommand(ByRef tbpEstimatedPose As typeBotPose, ByRef tbpTargetPose As typeBotPose, ByRef strCommand As String, ByRef tbpStartingPose As typeBotPose) As typeBotPose
        Try
            ftnExecuteMoveCommand = tbpEstimatedPose
            boolMoveComplete = False
            If boolDebugDrive Then frmDebug.subWriteLogs(Color.Green, ftnElapsedTime(timeMoveCommandSent) & " ftnExecuteMoveCommand Start:  " & strCommand)
            boolCollisionPredicted = False
            strStatusDriving = intDrivingSequenceNumber & " : Driving: " & strCommand
            If Not boolBotIsMoving Then ftnCheckBatteryForDisconnect() '   
            tbpBotPose = tbpEstimatedPose
            timeMoveCommandSent = Now()
            tbpStartingPose = tbpEstimatedPose
            Dim strResponse As String = ftnSendCommand(strCommand)
            ftnExecuteMoveCommand = ftnCheckForMoveComError(ftnExecuteMoveCommand, strResponse, tbpTargetPose)
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnExecuteMoveCommand")
        End Try
    End Function


    Public Function ftnPostMovement(ByRef tbpEstimatedPose As typeBotPose, ByRef strTypeOfMove As String, _
                                    ByRef tbpTargetPose As typeBotPose) As typeBotPose
        'localizes, finds new path if required, collision recovery if required
        Try
            boolCollisionPredicted = False
            ftnPostMovement = tbpEstimatedPose
            If Not boolStop Then
                If boolDebugDrive Then frmDebug.subWriteLogs(Color.Red, ftnElapsedTime(timeMoveCommandSent) & " ftnPostMovement   START")
                ftnPostMovement = tbpEstimatedPose
                If boolSimulatedMovement Then
                    If (boolCollisionDetected) Then
                        ftnPostMovement = ftnCollisionRecovery(ftnPostMovement)
                    Else
                        If (intPathArray(ftnPostMovement.Location.X, ftnPostMovement.Location.Y) < constDestinationCost) Then  ' only required if the Bot is too far off the path
                            subGeneratePathArray()
                            ftnGeneratePathCosts(tmcDestination, ftnPostMovement.Location)
                        End If
                    End If
                ElseIf Not boolShutDown Then
                    If (tbpEstimatedPose.OrientationUncertainty < 60) Then ftnPostMovement = frmCompass.ftnLocalizeUsingCompass(ftnPostMovement) ' Don't use Compass during Lost Localization
                    Dim boolGoodLocalization As Boolean = False
                    If (boolCollisionDetected) Then
                        ftnPostMovement = ftnLocalize_Top(ftnPostMovement, boolGoodLocalization)
                        ftnPostMovement = ftnCollisionRecovery(ftnPostMovement)
                        boolRangeReadingsAreFresh = False
                    Else
                        boolRangeReadingsAreFresh = False
                        If ftnLocalizationNeeded(tbpEstimatedPose) Then ' Fewer localizations result in fewer driving corrections, but also greater chance of collisions - its a tradeoff
                            ftnPostMovement = ftnLocalize_Top(ftnPostMovement, boolGoodLocalization)
                            If boolGoodLocalization Then tbpLastLocalizedPose = ftnPostMovement
                        ElseIf boolSlam Then
                            subGeneratePathArray()
                        End If
                    End If
                    'If ftnAmIStuck(ftnPostMovement, strTypeOfMove, ftnAngleDifference(tbpEstimatedPose.Orientation, tbpTargetPose.Orientation)) Then _
                    '    ftnPostMovement = ftnGetUnStuck(ftnPostMovement, strTypeOfMove)
                    ftnPostMovement = ftnLostRecoveryTop(ftnPostMovement)
                    If (intPathArray(ftnPostMovement.Location.X, ftnPostMovement.Location.Y) < constDestinationCost) Then
                        subGeneratePathArray() ' Bot is too far off the original path, must regenerate path
                        ftnGeneratePathCosts(tmcDestination, ftnPostMovement.Location)
                    End If
                    tbpTargetPoseGlobal = ftnPostMovement
                End If
                If boolDebugDrive Then frmDebug.subWriteLogs(Color.Red, ftnElapsedTime(timeMoveCommandSent) & " ftnPostMovement   END" & vbCrLf & vbCrLf & vbCrLf & vbCrLf)
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnPostMovement")
        End Try
    End Function

    Function ftnCheckForMoveComError(ByRef tbpEstimatedPose As typeBotPose, ByRef strResponse As String, ByRef tbpTargetPose As typeBotPose) As typeBotPose
        Try
            ftnCheckForMoveComError = tbpEstimatedPose
            If CBool(InStr(strResponse, "COM ERROR")) Then ' some type of error was detected so increase uncertainties
                If (String.CompareOrdinal(strTypeOfMove, "Fwd") = 0) Or (String.CompareOrdinal(strTypeOfMove, "XY") = 0) Then
                    Dim sngDistanceToTarget As Single = ftnDistanceBetweenLocations(tbpEstimatedPose.Location, tbpTargetPose.Location)
                    ftnCheckForMoveComError.LocationUncertainty.X += sngDistanceToTarget / 2 / sngGridSize
                    ftnCheckForMoveComError.LocationUncertainty.Y += sngDistanceToTarget / 2 / sngGridSize
                ElseIf (String.CompareOrdinal(strTypeOfMove, "Turn") = 0) Or (String.CompareOrdinal(strTypeOfMove, "XY") = 0) Then
                    ftnCheckForMoveComError.OrientationUncertainty += Abs(ftnAngleDifference(tbpEstimatedPose.Orientation, tbpTargetPose.Orientation) / 2)
                End If
                ftnSendCommand("play phrase 'Communication error.  '")
                intMovementsSinceLocalization = 99  ' force compass and localization updates
                frmDebug.subWriteLogs(Color.Red, "  ")
                frmDebug.subWriteLogs(Color.Red, ftnElapsedTime(timeMoveCommandSent) & " **************  Communication Error during movement.  ' *****************************")
                frmDebug.subWriteLogs(Color.Red, "  ")
            ElseIf CBool(InStr(strResponse, "com timeout")) Then ' some type of error was detected so increase uncertainties
                ftnCheckForMoveComError = tbpTargetPose
                ftnCheckForMoveComError.LocationUncertainty.X = Max(ftnCheckForMoveComError.LocationUncertainty.X, 12)
                ftnCheckForMoveComError.LocationUncertainty.Y = Max(ftnCheckForMoveComError.LocationUncertainty.Y, 12)
                ftnCheckForMoveComError.OrientationUncertainty = Max(ftnCheckForMoveComError.OrientationUncertainty, 17)
                intMovementsSinceLocalization = 99  ' force compass and localization updates
                frmDebug.subWriteLogs(Color.Red, "  ")
                frmDebug.subWriteLogs(Color.Red, ftnElapsedTime(timeMoveCommandSent) & " **************  Communication Timeout during movement.  ' *****************************")
                frmDebug.subWriteLogs(Color.Red, "  ")
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnCheckForMoveComError")
        End Try
    End Function

    Function ftnSimulatedExecuteMove(ByVal tbpEstimatedPose As typeBotPose, ByVal strCommand As String, ByVal tbpTargetPose As typeBotPose) As typeBotPose
        On Error Resume Next
        ftnSimulatedExecuteMove = tbpEstimatedPose
        tbpTargetPoseGlobal = tbpTargetPose
        frmDebug.subDebugLog(Color.Black, ">: " & strCommand)
        Dim sngStartOfMoveDelayOld As Single = sngStartOfMoveDelay
        sngStartOfMoveDelay = 0
        subDoEventsIfMainThread()
        timeMoveCommandSent = Now()
        If (String.CompareOrdinal(strTypeOfMove, "stop") = 0) Then
            tbpTargetPoseGlobal = tbpEstimatedPose
            strTypeOfMove = ""
        Else
            tbpStartingPose = tbpEstimatedPose
            Dim sngDistanceToTarget As Single = ftnDistanceBetweenLocations(tbpEstimatedPose.Location, tbpTargetPose.Location)
            frmDebug.subWriteMoveStartLogFile(strCommand, tbpTargetPose)
            While ((String.CompareOrdinal(strTypeOfMove, "Turn") = 0) And ((Abs(ftnConvertTo180(ftnAngleDifference(tbpEstimatedPose.Orientation, tbpTargetPose.Orientation))) - 0.5) >= Abs(ftnAngleDifference(tbpStartingPose.Orientation, ftnSimulatedExecuteMove.Orientation))) Or _
                 (String.CompareOrdinal(strTypeOfMove, "Fwd") = 0) And ((Abs(sngDistanceToTarget) - sngGridSize / 2) >= ftnDistanceBetweenLocations(tbpStartingPose.Location, ftnSimulatedExecuteMove.Location)) Or _
                 (String.CompareOrdinal(strTypeOfMove, "XY") = 0) And ((Abs(sngDistanceToTarget) - sngGridSize / 2) >= ftnDistanceBetweenLocations(tbpStartingPose.Location, ftnSimulatedExecuteMove.Location))) And _
                 Not boolStop And Not boolShutDown And Not ftnPosesEqual(tbpEstimatedPose, tbpTargetPose)
                boolBotIsMoving = True
                Dim tbpOldLocation As typeMapCoordinate = ftnSimulatedExecuteMove.Location
                ftnSimulatedExecuteMove = ftnLocalizeUsingDeadReckoning(ftnSimulatedExecuteMove, 0, timeMoveCommandSent, tbpStartingPose, tbpTargetPose, strTypeOfMove, False)
                tbpBotPose = ftnSimulatedExecuteMove
                If Not ftnLocationsEqual(ftnSimulatedExecuteMove.Location, tbpOldLocation) Then subEraseOldBotLocation(tbpOldLocation)
                subDrawBot(ftnSimulatedExecuteMove)
                subDoEventsIfMainThread()
            End While
        End If
        'tbpStartingPose = ftnSimulatedExecuteMove
        intDrivingSequenceNumber += 1
        tbpTargetPoseGlobal = ftnSimulatedExecuteMove
        boolBotIsMoving = False
        boolCollisionPredicted = False
        sngStartOfMoveDelay = sngStartOfMoveDelayOld
    End Function



    Public Function ftnLocalizationNeeded(ByRef tbpEstimatedPose As typeBotPose) As Boolean
        Try
            ftnLocalizationNeeded = False
            If boolSimulatedMovement Or boolShutDown Then
                ftnLocalizationNeeded = False
            ElseIf (ftnPoseUncertaintyFactor(tbpEstimatedPose) > 480) Then '(15+15)*16
                ftnLocalizationNeeded = True
            ElseIf Not boolPoseConfirmedAfterCollision Then
                ftnLocalizationNeeded = True
            ElseIf (Abs(ftnAngleDifference(tbpLastLocalizedPose.Orientation, tbpEstimatedPose.Orientation)) >= 25) Then
                ftnLocalizationNeeded = True
            ElseIf (ftnDistanceBetweenLocations(tbpLastLocalizedPose.Location, tbpEstimatedPose.Location) > sngSixInches) Then
                ftnLocalizationNeeded = True
            ElseIf boolSlam Or boolEnableRBPF Then
                ftnLocalizationNeeded = True '(Abs(ftnAngleDifference(tbpEstimatedPose.Orientation, tbpLastLocalizedPose.Orientation)) > 10)  '  Restrict endless back and forth turns
            ElseIf (intMovementsSinceLocalization > 9) Then
                ftnLocalizationNeeded = True
            ElseIf (intConsecutiveTurnCounter <= constMaxTurnsWithLocalization) Then ' This prevents Bot from turning back and forth repeatedly
                ftnLocalizationNeeded = (intMovementsSinceLocalization > 2)
            End If
            If ftnLocalizationNeeded Then
                intMovementsSinceLocalization = 0
            Else
                intMovementsSinceLocalization += 1
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnLocalizationNeeded")
        End Try
    End Function


    Function ftnCheckBatteryForDisconnect() As Boolean
        Try
            ftnCheckBatteryForDisconnect = False
            If Not boolSimulatedMovement Then
                Do
                    ftnSendCommand("battery") '   
                    If (strStatusBotVoltage = "   ??   ") Or CBool(InStr(strStatusBotVoltage, "--")) Or (sngBotBatteryVoltage < (sngDockTriggerBot * 0.7)) Then _
                        ftnSendCommand("battery") '   recheck in case of comm error
                    If (strStatusBotVoltage = "   ??   ") Or CBool(InStr(strStatusBotVoltage, "--")) Or (sngBotBatteryVoltage < (sngDockTriggerBot * 0.7)) Then
                        ftnCheckBatteryForDisconnect = True
                        GotABotClass.StatusDriving.ForeColor = Color.Red
                        strStatusDriving = intDrivingSequenceNumber & " : BATTERY DISCONNECTED! "
                        frmDebug.subWriteLogs(Color.Black, Now() & "    |||||||||||  BATTERY DISCONNECT DETECTED  |||||||||||")
                        frmDebug.subWriteLogs(Color.Black, Now() & "                   Battery Voltage = " & sngBotBatteryVoltage)
                        GotABotClass.subRedraw(tbpBotPose)
                                    ftnSendCommand("play phrase Beep")
                        subWait(1500)
                        ftnSendCommand("battery") '   
                    End If
                Loop Until (strStatusBotVoltage <> "   ??   ") Or Not CBool(InStr(strStatusBotVoltage, "--")) Or (sngBotBatteryVoltage > (sngDockTriggerBot * 0.7)) Or boolStop Or boolShutDown
                If ftnCheckBatteryForDisconnect Then
                    frmDebug.subWriteLogs(Color.Black, Now() & "                   Battery Reconnected.  Voltage = " & sngBotBatteryVoltage)
                End If
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnCheckBatteryForDisconnect")
        End Try
    End Function

    Function ftnAmIStuck(ByRef tbpEstimatedPose As typeBotPose, ByRef strTypeOfMove As String, ByRef sngMoveDistance As Single) As Boolean
        On Error Resume Next
        ftnAmIStuck = False
        If Not boolEnableRBPF And Not boolSlam And (String.CompareOrdinal(strTypeOfMove, "Turn") = 0) Then
            If (Abs(ftnAngleDifference(tbpTargetPoseGlobal.Orientation, tbpEstimatedPose.Orientation)) > 20) And Not boolStop Then
                ftnAmIStuck = True
                ftnSendCommand("play phraseA 'I am stuck'")
                frmDebug.subWriteLogs(Color.Black, Now() & "    |||||||||||  I'm stuck  |||||||||||")
            End If
        End If
    End Function

    Public Function ftnGetUnStuck(ByRef tbpEstimatedPose As typeBotPose, ByRef strTypeOfMove As String) As typeBotPose
        On Error Resume Next
        Dim boolGoodLocalization As Boolean = False
        ftnGetUnStuck = tbpEstimatedPose
        If Not boolStop And Not boolShutDown And boolDrivingToDestination And Not (GotABotClass.picStop.BorderStyle = FormBorderStyle.Fixed3D) Then
            subDoEventsIfMainThread()
            If (String.CompareOrdinal(strTypeOfMove, "Turn") = 0) Then
                tbpTargetPoseGlobal.Location = ftnLocationFromEstimatedPose(tbpBotPose, 0, sngSixInches)
                tbpBotPose = ftnBotFwd(tbpBotPose, tbpTargetPoseGlobal)
                ftnGetUnStuck = ftnLocalize_Top(tbpBotPose, boolGoodLocalization)
                tbpBotPose = ftnGetUnStuck
                tbpTargetPoseGlobal.Location = ftnLocationFromEstimatedPose(tbpBotPose, 0, -sngTwoInches)
                ftnGetUnStuck = ftnBotFwd(tbpBotPose, tbpTargetPoseGlobal)
            ElseIf (String.CompareOrdinal(strTypeOfMove, "Fwd") = 0) Then
                tbpTargetPoseGlobal.Location = ftnLocationFromEstimatedPose(tbpBotPose, 0, -sngSixInches)
                tbpBotPose = ftnBotFwd(tbpBotPose, tbpTargetPoseGlobal)
                ftnGetUnStuck = ftnLocalize_Top(tbpBotPose, boolGoodLocalization)
                tbpBotPose = ftnGetUnStuck
                tbpTargetPoseGlobal.Location = ftnLocationFromEstimatedPose(tbpBotPose, 0, sngTwoInches)
                ftnGetUnStuck = ftnBotFwd(tbpBotPose, tbpTargetPoseGlobal)
            End If
            ftnGetUnStuck = ftnLocalize_Top(tbpBotPose, boolGoodLocalization)
            subDoEventsIfMainThread()
        End If
        tbpBotPose = ftnGetUnStuck
    End Function

End Module
