﻿
Imports System.Math

Module modDriveToDestination

    ' This module determines the driving commands required for the Bot
    ' to reach a specific destination

    ' Examples of  driving commands:
    '      move 27 inches
    '      move -12 cm
    '      move 27 degrees
    '      move x 13 y 8 cm
    '      stop


    Public Sub subDriveToUserChosenDestination(ByVal tmcNewDestination As typeMapCoordinate)
        Try
            tmcDestination = tmcNewDestination  '
            GotABotClass.subRedraw(tbpBotPose)
            subDrawDestination(tmcDestination)
            If ftnInWall(tmcNewDestination) Then
                strStatusDriving = "In Wall"
                ftnSendCommand("play phrase ' I'm sorry Dave. I'm afraid I can't do that '")
            Else
                boolStop = False
                boolDock = False
                GotABotClass.picStop.BorderStyle = FormBorderStyle.None
                frmDebug.subWriteLogs(Color.Black, " ")
                frmDebug.subWriteLogs(Color.Black, " ")
                frmDebug.subWriteLogs(Color.Black, " ")
                frmDebug.subWriteLogs(Color.Black, " ******************************************************************************************")
                frmDebug.subWriteLogs(Color.Black, " ")
                frmDebug.subWriteLogs(Color.Black, " GOTABOT       (Version " & My.Application.Info.Version.ToString(3) & ")")
                ftnSendCommand("play phraseA OK")
                ftnDriveToDestination_Top(tbpBotPose, tmcNewDestination) ' Drive to destination
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in subDriveToUserChosenDestination " & ex.ToString)
        End Try
    End Sub
    Public Function ftnDriveToDestination_Top(ByVal tbpStartingPose As typeBotPose, ByVal tmcDestination As typeMapCoordinate) As Boolean
        Try
            ftnDriveToDestination_Top = False
            frmDebug.subWriteLogs(Color.Black, Now() & " Drive to new destination: " & tmcDestination.X & ", " & tmcDestination.Y)
            strStatusDriving = "Drive to " & tmcDestination.X & ", " & tmcDestination.Y
            If boolBotIsMoving Then
                tbpBotPose = ftnLocalizeUsingDeadReckoning(tbpBotPose, 0, timeMoveCommandSent, tbpStartingPose, tbpTargetPoseGlobal, strTypeOfMove, True)
                tbpBotPose = ftnBotStop(tbpBotPose, tbpTargetPoseGlobal)
            End If
            subInitializeDrivingControl(tbpBotPose, tbpStartingPose)  ' initialize
            intLastCollisionSequenceNumber = -6
            boolCollisionDetectionEnable = True
            If ftnCheckForStopMode() Then Exit Function
            subCheckForPause()
            timeDrivingStarted = Now()
            boolDrivingToDestination = True
            boolRangeReadingsAreFresh = False
            boolCollisionDetected = False
            boolCollisionPredicted = False
            ftnSendCommand("stop")
            intDrivingSequenceNumber = 0
            frmDock.subUnDock(tbpStartingPose)
            subDrawDestination(tmcDestination)
            If boolMapUpdatedWithObstacle Then
                subConvertMapToArray()
                boolMapUpdatedWithObstacle = False
            End If
            subGeneratePathArray()
            ftnGeneratePathCosts(tmcDestination, tbpBotPose.Location)
            subDriveToDestinationLoop(tbpStartingPose, tmcDestination)
            If Not GotABotClass.mnuDock.Checked And Not GotABotClass.InvokeRequired And Not boolDock Then
                frmDebug.subWriteFinishedMessage()
                GotABotClass.StatusDriving.ForeColor = Color.Black
                strStatusDriving = intDrivingSequenceNumber & " : Finished Driving "
                ftnDriveToDestination_Top = (ftnDistanceBetweenLocations(tbpBotPose.Location, tmcDestination) <= sngBotDiameter / 3)
                If Not boolSlam And (ftnDistanceBetweenLocations(tbpBotPose.Location, tmcDestination) <= (sngBotDiameter / 6)) Then
                    ftnSendCommand("play phraseA 'nailed it.'")
                ElseIf Not boolSlam And ftnDriveToDestination_Top Then
                    ftnSendCommand("play phraseA 'close enough'")
                Else
                    ftnSendCommand("play phrase Beep")
                End If
            End If
            GotABotClass.subRedraw(tbpBotPose)
            boolDrivingToDestination = False
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnDriveToDestination_Top")
        End Try
    End Function

    Sub subDriveToDestinationLoop(ByVal tbpStartingPose As typeBotPose, ByVal tmcDestination As typeMapCoordinate)
        Try
            Dim boolGoodPath As Boolean = True
            Dim tbpTargetPose As typeBotPose
            Dim sngDestinationError As Single
            If boolSlam Then
                sngDestinationError = sngBotDiameter * 0.5
            Else
                sngDestinationError = sngBotDiameter * 0.33333
            End If
            Do While ((ftnDistanceBetweenLocations(tbpBotPose.Location, tmcDestination) > sngDestinationError) _
                             And Not boolStop And boolGoodPath And boolDrivingToDestination)
                If boolDebugDrive Then frmDebug.subWriteLogs(Color.Red, ftnElapsedTime(timeMoveCommandSent) & "          <<<<<  subDriveToDestinationLoop  START      ")
                subCheckForPause()
                GotABotClass.subRedraw(tbpBotPose)

                If Not GotABotClass.InvokeRequired Then GotABotClass.StatusDriving.ForeColor = Color.Black
                If boolDebugDrive Then frmDebug.subWritePose(ftnElapsedTime(timeMoveCommandSent) & " subDriveToDestinationLoop: " & " tbpBotPose  ", tbpBotPose)
                tbpTargetPose = ftnFindTargetPose(tbpBotPose, tmcDestination) ' 
                If boolDebugDrive Then frmDebug.subWritePose(ftnElapsedTime(timeMoveCommandSent) & " subDriveToDestinationLoop: " & "  Target Pose  ", tbpTargetPose)
                tbpBotPose = ftnControlDriving(tbpStartingPose, tbpBotPose, tbpTargetPose, strTypeOfMove)  '  
                If Not (boolSlam And Not boolDrivingToDestination) Then tbpBotPose = ftnPostMovement(tbpBotPose, strTypeOfMove, tbpTargetPoseGlobal)
                If (ftnPoseUncertaintyFactor(tbpBotPose) < 288) Then boolPoseConfirmedAfterCollision = True '(288 = (12 inches + 12 inches) * 12 degrees) _
                strTypeOfMove = ""
            Loop  ' 
            If boolDebugDrive Then frmDebug.subWriteLogs(Color.Red, ftnElapsedTime(timeMoveCommandSent) & "          <<<<<  subDriveToDestinationLoop  END     ")
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in subDriveToDestinationLoop")
        End Try
    End Sub


    Function ftnControlDriving(ByVal tbpStartingPose As typeBotPose, ByVal tbpEstimatedPose As typeBotPose, ByVal tbpTargetPose As typeBotPose, ByRef strTypeOfMove As String) As typeBotPose
        Try
            ftnControlDriving = tbpEstimatedPose
            If boolDebugDrive And Not boolBotIsMoving Then frmDebug.subWriteLogs(Color.Red, ftnElapsedTime(timeMoveCommandSent) & " ftnControlDriving  START  ")
            tbpTargetPoseGlobal = tbpTargetPose
            Dim intTurnAngle As Short = ftnTurnAngle(tbpEstimatedPose, tbpTargetPose)
            strTypeOfMove = ftnTypeOfMove(tbpEstimatedPose, tbpTargetPose)
            If boolStop Then
                ftnControlDriving = ftnLocalizeUsingDeadReckoning(tbpEstimatedPose, 0, timeMoveCommandSent, tbpStartingPose, tbpTargetPose, strTypeOfMove, True)
                ftnControlDriving = ftnBotStop(ftnControlDriving, tbpTargetPose)
            ElseIf Not boolBotIsMoving And ((String.CompareOrdinal(strTypeOfMove, "Fwd") = 0)) And ftnObstacleDetected(tbpEstimatedPose, strTypeOfMove, tbpTargetPoseGlobal) Then  ' TODO Add Or (String.CompareOrdinal(strTypeOfMove, "XY") = 0)
                If boolSlam Then
                    ftnControlDriving = tbpEstimatedPose
                    boolDrivingToDestination = False
                Else
                    ftnControlDriving = ftnObstacleAvoid(tbpEstimatedPose, tbpTargetPoseGlobal)
                End If
            ElseIf Not boolStop Then
                Select Case strTypeOfMove
                    Case "Turn"
                        ' TODO - move back up function to modNavigationPath 
                        Dim sngDistanceToTarget As Single = ftnDistanceBetweenLocations(tbpEstimatedPose.Location, tbpTargetPose.Location)
                        If (Abs(intTurnAngle) < 135) Or (sngDistanceToTarget > sngTenInches) Or boolSlam Then
                            tbpTargetPoseGlobal = tbpEstimatedPose
                            tbpTargetPoseGlobal.Orientation = ftnConvertTo360(ftnControlDriving.Orientation + intTurnAngle)
                            ftnControlDriving = ftnBotTurn(ftnControlDriving, True, tbpTargetPoseGlobal)
                        Else
                            frmDebug.subWriteLogs(Color.Red, ftnElapsedTime(timeMoveCommandSent) & " ***** Turn and Backup: . ***************************")
                            tbpTargetPoseGlobal.Orientation = ftnConvertTo360(ftnControlDriving.Orientation + intTurnAngle - 180)
                            ftnControlDriving = ftnBotTurn(ftnControlDriving, False, tbpTargetPoseGlobal)  ' prepare to backup
                            boolPoseConfirmedAfterCollision = False  ' force localization
                            ftnControlDriving = ftnBotFwd(ftnControlDriving, tbpTargetPoseGlobal) ' backup
                        End If
                    Case "Fwd"
                        ftnControlDriving = ftnBotFwd(ftnControlDriving, tbpTargetPoseGlobal)
                    Case "XY"
                        tbpTargetPoseGlobal = ftnXY_DR(tbpEstimatedPose, 99999, tbpTargetPose, False) ' Calculate the orientation
                        ftnControlDriving = ftnBotXY(ftnControlDriving, tbpTargetPoseGlobal)
                    Case "Stop"
                        ftnControlDriving = ftnLocalizeUsingDeadReckoning(ftnControlDriving, 0, timeMoveCommandSent, tbpStartingPose, tbpTargetPose, strTypeOfMove, True)
                        ftnControlDriving = ftnBotStop(ftnControlDriving, tbpTargetPoseGlobal)
                    Case Else
                        frmDebug.subWriteLogs(Color.Red, " **********************  ftnControlDriving:  Unexpected strTypeOfMove")
                        ftnControlDriving = ftnLocalizeUsingDeadReckoning(ftnControlDriving, 0, timeMoveCommandSent, tbpStartingPose, tbpTargetPose, strTypeOfMove, True)
                        ftnControlDriving = ftnBotStop(ftnControlDriving, tbpTargetPoseGlobal)
                End Select
            End If
            If boolDebugDrive And Not boolBotIsMoving Then frmDebug.subWriteLogs(Color.Red, ftnElapsedTime(timeMoveCommandSent) & " ftnControlDriving  END     ")
            subDoEventsIfMainThread()
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnControlDriving")
        End Try
    End Function



End Module
