﻿Imports System
Imports System.Math
Imports System.Net
Imports System.Threading
Imports System.IO.Ports


'.  Pilot Motion Processor command
'            1st byte: address
'            2nd byte: checksum
'            3nd byte: axis (always 00)
'            4th byte: name of command (command code)
'            5th byte+: parameters of the command. (higher order bytes send first)
'



Public Structure PWDReply
    Dim data() As Byte  'data[PWMREPLYSIZE]
    Dim dwSize As Integer
    Dim Status() As Integer ' return 0(valid) if the valus is correct
    Dim CheckChecksum() As Integer
    '	void PrintData();
End Structure


Module modIF_ER1_Serial


    '==================================  Serial Interface  =====================================================
    Private comBot As New SerialPort

    Public Function ftnER1_SerialConfig() As Boolean
        ' returns true if comBot is successfully opened
        Try
            ftnER1_SerialConfig = False
            If boolDebugER1_Serial Then frmDebug.subWriteLogs(Color.Green, " ftnER1_SerialConfig Start ")
            If Not boolBotConnected Then
                Dim availableSerialPorts As System.Collections.ObjectModel.ReadOnlyCollection(Of String) = My.Computer.Ports.SerialPortNames
                If Not boolShutDown And (strBotComPort <> "None") And availableSerialPorts.Contains(strBotComPort) Then
                    If Not comBot.IsOpen Then
                        With comBot
                            .PortName = strBotComPort
                            .BaudRate = 250000
                            .DataBits = 8
                            .Parity = Parity.None
                            .StopBits = StopBits.One
                            .Handshake = Handshake.None
                        End With
                        comBot.Open()
                    End If
                    If Not comBot.IsOpen Then
                        ftnSendCommand("play phrase Beep")
                    Else
                        For i = 1 To 4
                            subWait(30)
                            boolBotConnected = True
                            comBot.ReadExisting()
                            boolBotConnected = ftnSendER1SerialCommand(4, 1, PilotSet.GetChecksum, 0, 0, 0, 0, 0, 0, 4) = &H12345678
                            If boolBotConnected Then _
                                boolBotConnected = ftnSendER1SerialCommand(4, 0, PilotSet.GetChecksum, 0, 0, 0, 0, 0, 0, 4) = &H12345678
                            If boolBotConnected Then Exit For
                        Next
                        If boolBotConnected Then
                            subResetWheels(PLProfile.Trapezoidal)
                            ftnER1_SerialConfig = True
                            If boolDebugER1_Serial Then frmDebug.subWriteLogs(Color.Green, " comBot configured ")
                        End If
                    End If
                End If
            End If
        Catch ex As Exception
            strStatusDriving = "ERROR: Robot not detected."
            ftnSendCommand("play phrase Beep")
            ftnER1_SerialConfig = False
            boolBotConnected = False
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnER1_SerialConfig " & ex.ToString)
        End Try
    End Function


    Public Function ftnSendER1SerialCommand(ByVal intBytesToSend As Short, ByVal byte1 As Byte, ByVal byte2 As Byte, ByVal byte3 As Byte _
                            , ByVal byte4 As Byte, ByVal byte5 As Byte, ByVal byte6 As Byte, ByVal byte7 As Byte, ByVal byte8 As Byte, _
                            ByRef intBytesToReceive As Integer) As UInt32
        ftnSendER1SerialCommand = 0
        Try
            subWait(10)
            If Not boolBotConnected Then
                strStatusDriving = "ER1 Not Connected "
                GotABotClass.subRedraw(tbpBotPose)
                boolBotConnected = ftnER1_SerialConfig()
            End If
            If boolBotConnected Then
                If (comBot.BytesToRead > 0) Then
                    Dim strUnread As String = ""
                    While (comBot.BytesToRead > 0)
                        strUnread &= Hex(comBot.ReadByte)
                    End While
                    If boolDebugER1_Serial And (strUnread <> "00") Then frmDebug.subWriteLogs(Color.Red, " Leftover data from previous command: " & strUnread)
                End If
                Dim byteBuffer As Byte() = { _
                                     byte1, _
                                     Convert.ToByte(((Not (byte1 + byte2 + byte3 + byte4 + byte5 + byte6 + byte7 + byte8) And &HFF) + 1) And &HFF), _
                                     0, _
                                     byte2, _
                                     byte3, _
                                     byte4, _
                                     byte5, _
                                     byte6, _
                                     byte7, _
                                     byte8 _
                                     }
                If boolDebugER1_Serial Then
                    Dim strSent As String = " }}}}}}}}}}====     Sent To ER1 Serial IF: "
                    For i = 0 To intBytesToSend - 1
                        strSent &= " " & Hex(byteBuffer(i))
                    Next
                    frmDebug.subWriteLogs(Color.Purple, strSent)
                End If
                comBot.Write(byteBuffer, 0, intBytesToSend)
                If (intBytesToReceive > 0) Then
                    ftnSendER1SerialCommand = ftnGetResponse(intBytesToReceive)
                    If boolDebugER1_Serial Then frmDebug.subWriteLogs(Color.Purple, " }}}}}}}}}}====       Received From ER1 Serial IF: " & Hex(ftnSendER1SerialCommand))
                End If
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnSendER1SerialCommand " & ex.ToString)
        End Try
    End Function

    Function ftnGetResponse(ByRef intBytesToReceive As Integer) As UInt32  '
        ftnGetResponse = 0
        Try
            Dim intBytesReceived As Byte = 0
            Dim timeOut As DateTime = Now().AddMilliseconds(1000)
            subWait(10)
            Do
                If comBot.BytesToRead > 0 Then ' read and discard status and checksum
                    ftnGetResponse = (ftnGetResponse << 8) + comBot.ReadByte
                    intBytesReceived += 1
                End If
            Loop Until (Now() > timeOut) Or (intBytesReceived = 2) Or boolShutDown
            If (ftnGetResponse > 255) Then frmDebug.subWriteLogs(Color.Red, " COM ERROR:  status / checksum = " & Hex(ftnGetResponse))
            intBytesReceived = 0
            ftnGetResponse = 0
            Do
                If comBot.BytesToRead > 0 Then
                    ftnGetResponse = (ftnGetResponse << 8) + comBot.ReadByte
                    intBytesReceived += 1
                End If
            Loop Until (Now() > timeOut) Or (intBytesToReceive = intBytesReceived) Or boolShutDown
            If (intBytesReceived <> intBytesToReceive) Then
                frmDebug.subWriteLogs(Color.Purple, ftnElapsedTime(timeMoveCommandSent) & "  ERROR:  only " & intBytesReceived & " of " & intBytesToReceive & " Bytes received")
                ftnGetResponse = 0
            End If
            If (comBot.BytesToRead > 0) Then
                subWait(10)
                Dim strUnread As String = ""
                While (comBot.BytesToRead > 0)
                    strUnread &= Hex(comBot.ReadByte)
                End While
                If boolDebugER1_Serial And (strUnread <> "00") Then frmDebug.subWriteLogs(Color.Red, " Post Message Data: " & strUnread)
            End If
        Catch ex As Exception
            ftnGetResponse = 0
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnGetResponse " & ex.ToString)
        End Try
    End Function


 
    Public Sub subCloseComBot()
        Try
            If comBot IsNot Nothing Then
                If comBot.IsOpen Then comBot.Close()
                comBot = Nothing
            End If
            boolBotConnected = False
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in subCloseComBot")
        End Try
    End Sub



End Module
