﻿'https://msdn.microsoft.com/en-us/library/7ya7y41k.aspx?f=255&MSPPError=-2147217396
Imports System
Imports System.Math
Imports System.Net
Imports System.Threading
Imports System.IO.Ports


Module modIF_Hero

    '================================  Hero Serial IF  =========================================
    Private comBot As New SerialPort
    Dim boolMoveInProgress As Boolean = False

    Public Function ftnHeroConfig() As Boolean
        ' returns true if comBot is successfully opened
        Try
            ftnHeroConfig = False
            If Not boolBotConnected Then
                Dim availableSerialPorts As System.Collections.ObjectModel.ReadOnlyCollection(Of String) = My.Computer.Ports.SerialPortNames
                If Not boolShutDown And (strBotComPort <> "None") And availableSerialPorts.Contains(strBotComPort) Then
                    With comBot
                        .PortName = strBotComPort
                        .BaudRate = 9600
                        .DataBits = 7
                        .Parity = Parity.None
                        .StopBits = StopBits.One
                        .Handshake = Handshake.None
                    End With
                    comBot.Open()
                    If Not comBot.IsOpen Then
                        ftnSendCommand("play phrase Beep")
                        MsgBox(" ERROR:  Hero Robot not detected.")
                    Else
                        ftnHero_Communicate(" ", ">", 1000) ' Empty Hero's interface buffer
                        If Not CBool(InStr(ftnHero_Communicate("GV", ">", 1000), "GotAHero")) Then
                            ftnSendCommand("play phrase Beep")
                            MsgBox(" Download Hero's interface program by" & vbCrLf & _
                                   "   1. Press 'Reset' then Enter '3A' on Hero" & vbCrLf & _
                                   "   2. Press 'OK' below" & vbCrLf & _
                                   "   3. Wait for download to complete" & vbCrLf & _
                                    vbCrLf & _
                                   " (Hero will say 'Ready' if download is sucessful)")
                            subHeroS19Download()
                            ftnSendCommand("play phrase Beep")
                            MsgBox(" Start Hero's interface program by" & vbCrLf & _
                                   "   1. Enter 'AD0200' on Hero" & vbCrLf & _
                                   " (Hero will display 'GH-100.' if sucessful)" & _
                                     vbCrLf & _
                                  "   2. Press 'OK' below" & vbCrLf & _
                                   " (Hero will display '-HEro-.' if sucessful)")
                            ftnHero_Communicate(" ", ">", 1000) ' Empty Hero's interface buffer
                            If Not CBool(InStr(ftnHero_Communicate("GV", ">", 1000), "GotAHero")) Then
                                ftnSendCommand("play phrase Beep")
                                MsgBox(" Download Failed")
                            Else
                                ftnHeroConfig = True
                                ftnHero_Communicate("GD08374F051D88", ">", 1000) ' Write "-Hero-" to robot's display
                                'ftnHero_Communicate("GU0", ">", 3000) ' Disable Ultrasonic Ranging
                            End If
                        Else
                            ftnHeroConfig = True
                        End If
                    End If
                End If
                If Not boolURGConnected And Not boolKinectConnected Then
                    sngScanFOV = 350
                    intScanVectorsMaxRangeInches = 100
                    intScanVectorsMinRangeInches = 0
                    subCalculateScanBeamAngles()
                End If
            End If
        Catch ex As Exception
            strStatusDriving = "ERROR: Hero Robot not detected."
            ftnSendCommand("play phrase Beep")
            MsgBox(" ERROR:  Hero Robot not detected.")
            ftnHeroConfig = False
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnHeroConfig")
            frmDebug.subWriteLogs(Color.Red, "    ERROR:  " & ex.Message)
        End Try
    End Function

    Dim boolHeroComInProgress As Boolean = False
    Public Function ftnHero_Communicate(ByRef strCommand As String, ByRef strResponse As String, ByRef inttimeOut As Integer) As String
        ftnHero_Communicate = ""
        Try
            If Not boolShutDown And Not boolHeroComInProgress Then
                boolHeroComInProgress = True
                If (comBot Is Nothing) Then
                    ftnHeroConfig()
                ElseIf Not comBot.IsOpen Then
                    ftnHeroConfig()
                End If
                If (comBot IsNot Nothing) Then
                    If comBot.IsOpen Then
                        comBot.ReadTimeout = Min(1000, inttimeOut)
                        boolBotConnected = True
                        subSendSerialData(strCommand)
                        If (strResponse <> "") Then
                            subWait(50)
                            ftnHero_Communicate = ftnWaitForExpectedResponse(strResponse, inttimeOut)
                            If boolDebugHero Then frmDebug.subWriteLogs(Color.Purple, " }}}}}}}}}}====     Received From Hero: " & ftnHero_Communicate)
                            ftnHero_Communicate = Replace(ftnHero_Communicate, strCommand, "", 1, -1)
                            ftnHero_Communicate = Replace(ftnHero_Communicate, vbCr, "", 1, -1)
                            ftnHero_Communicate = Replace(ftnHero_Communicate, vbLf, "", 1, -1)
                            ftnHero_Communicate = Replace(ftnHero_Communicate, " ", "", 1, -1)
                            ftnHero_Communicate = Replace(ftnHero_Communicate, strResponse, "", 1, -1)
                        End If
                    End If
                End If
                boolHeroComInProgress = False
            End If
        Catch ex As Exception
            boolHeroComInProgress = False
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnHero_Communicate")
        End Try
    End Function

    Function ftnWaitForExpectedResponse(ByRef strTextToWaitFor As String, ByRef inttimeOut As Integer) As String  '
        ftnWaitForExpectedResponse = ""
        Try
            Dim timeOut As DateTime = Now().AddMilliseconds(inttimeOut)
            Do
                ftnWaitForExpectedResponse &= ftnReceiveSerialData()
                subUpdateMapDuringMove()
            Loop Until (Now() > timeOut) Or CBool(InStr(ftnWaitForExpectedResponse, strTextToWaitFor)) Or boolShutDown
            If (Now() > timeOut) Then
                frmDebug.subWriteLogs(Color.Purple, ftnElapsedTime(timeMoveCommandSent) & "  WARNING: Bot Communication Timeout while waiting for response: " & strTextToWaitFor & " /// Actual Response = " & ftnWaitForExpectedResponse)
                ftnWaitForExpectedResponse &= " com timeout"
            End If
        Catch ex As Exception
            'subCloseComBot()
            ftnWaitForExpectedResponse &= " COM ERROR"
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnWaitForExpectedResponse")
        End Try
    End Function


    Function ftnReceiveSerialData() As String
        ' Receive strings from a serial port. 
        Try
            ftnReceiveSerialData = comBot.ReadExisting()
        Catch ex As TimeoutException
            ftnReceiveSerialData = ""
            'subCloseComBot()
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnReceiveSerialData = " & ex.ToString)
        End Try
    End Function

    Sub subSendSerialData(ByVal strData As String)
        ' Send strings to a serial port. 
        Try
            Dim strLetter As String = ""
            For i As Short = 1 To Len(strData)
                strLetter = Mid(strData, i, 1)
                comBot.Write(strLetter)
                Thread.Sleep(5)
            Next
            comBot.Write(vbCrLf)
        Catch ex As Exception
            'subCloseComBot()
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in subSendSerialData")
        End Try
    End Sub

    Public Sub subCloseComBot()
        Try
            If comBot IsNot Nothing Then
                If comBot.IsOpen Then comBot.Close()
                comBot = Nothing
            End If
            boolBotConnected = False
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in subCloseComBot")
        End Try
    End Sub



    '================================  Hero S19 Download  =========================================

    Public Sub subHeroS19Download()
        ' http://www.homeandlearn.co.uk/NET/nets8p2.html
        Try
            If (Dir("GOTAHERO.S19") <> "") Then
                strStatusDriving = "Hero Downloading... "
                GotABotClass.subRedraw(tbpBotPose)
                Dim intLinesRead As Short = 0
                Dim intLinesInFile As Short = ftnLineCount("GOTAHERO.S19")
                Dim strTextLine As String = ""
                Dim objReader As New System.IO.StreamReader("GOTAHERO.S19")
                GotABotClass.pgbProgressBar1.Visible = True
                GotABotClass.pgbProgressBar1.Value = 0
                Do While objReader.Peek() <> -1
                    strTextLine = objReader.ReadLine() & vbNewLine
                    If boolDebugHero Then frmDebug.subWriteLogs(Color.Red, strTextLine)
                    intLinesRead += 1
                    GotABotClass.pgbProgressBar1.Value = intLinesRead / intLinesInFile * 100
                    ftnHero_Communicate(strTextLine, "", 500)
                    subWait(80)
                Loop
                objReader.Close()
                GotABotClass.pgbProgressBar1.Visible = False
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in subHeroS19Download")
        End Try
    End Sub

    Function ftnLineCount(ByVal strFileName As String) As Short
        Try
            ftnLineCount = 0
            Dim obj As New System.IO.StreamReader(strFileName)
            Do Until obj.ReadLine Is Nothing
                ftnLineCount += 1
            Loop
            obj.Close()
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnLineCount")
        End Try
    End Function


    '================================  Scan Vectors   =========================================
    Function ftnGetHeroScanVectors(ByRef tbpEstimatedPose As typeBotPose) As Boolean
        ' Rotate Hero's Head while taking sonar range measurements 
        Try
            strStatusDriving = "Measuring Scan Vectors... "
            GotABotClass.subRedraw(tbpBotPose)
            'ftnHero_Communicate("GU1", ">", 3000) ' Enable Ultrasonic Ranging
            If classScanner.chkContinuous.Checked Then ' don't rotate - 
                intHeadAngle = 0 + intScanOffsetDegrees
                subDrawVectorFromBot(tbpEstimatedPose, ftnGetHeroSonarRange(), intHeadAngle, True)
                subDrawBot(tbpEstimatedPose)
                System.Windows.Forms.Application.DoEvents()  '??
            Else
                intHeadAngle = trvScanVectors(0).sngAngle
                tbpTargetPoseGlobal = tbpBotPose   ' ??? 
                boolBotIsMoving = False   ' ??? 
                Dim boolVoiceCommandEnableTemp As Boolean = boolVoiceCommandEnable
                boolVoiceCommandEnable = False ' Noise from Head movement can cause false voice recognitions
                ftnHero_Communicate("GMC3C800H", ">", 90000) ' Move Head fully CCW
                Dim intHeadPosition As Short
                Dim strHeadPositionHex As String
                For intBeamNumber As Short = 0 To constMaxScannerBeams - 1 Step constMaxScannerBeams \ intNumberOfBeams
                    subCheckForPause()
                    If boolStop Then Exit For
                    If (intBeamNumber <> 0) Then intHeadPosition = intBeamNumber * 194 / (constMaxScannerBeams - 1)
                    strHeadPositionHex = Hex(intHeadPosition).ToString
                    If (Len(strHeadPositionHex) = 1) Then strHeadPositionHex = "0" & strHeadPositionHex
                    ftnHero_Communicate("GMC3CC" & strHeadPositionHex, ">", 3000) ' Move Head to intHeadPosition
                    intHeadAngle = trvScanVectors(intBeamNumber).sngAngle
                    trvScanVectors(intBeamNumber).sngRange = ftnGetHeroSonarRange()
                    subDrawVectorFromBot(tbpEstimatedPose, trvScanVectors(intBeamNumber).sngRange, trvScanVectors(intBeamNumber).sngAngle, Color.Indigo)
                    subDrawBot(tbpEstimatedPose)
                    strStatusDriving = "Measuring Scan Vectors: " & intBeamNumber & " / " & intNumberOfBeams
                    System.Windows.Forms.Application.DoEvents()  '??
                Next intBeamNumber
                'ftnHero_Communicate("GU0", ">", 3000) ' Disable Sonar
                If Not boolStop Then
                    intHeadAngle = 0 + intScanOffsetDegrees
                    ftnHero_Communicate("GMC3C861H", ">", 60000) ' Move Head CCW to center
                    subDrawBot(tbpEstimatedPose)
                End If
                boolVoiceCommandEnable = boolVoiceCommandEnableTemp
            End If
            ftnGetHeroScanVectors = True
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnGetHeroScanVectors - " & ex.ToString)
        End Try
    End Function


    Function ftnConvertHexStringToInteger(ByRef strHex As String, ByRef intDigits As Integer) As Integer
        Try
            ftnConvertHexStringToInteger = 0
            Dim intCharValue As Integer
            strHex &= "0000"
            While (ftnConvertHexCharToInteger(Left(strHex, 1)) < 0) And (Len(strHex) > intDigits)
                strHex = Replace(strHex, Left(strHex, 1), "", 1, 1) ' Get rid of any leading non Hex characters
            End While
            For i = 1 To intDigits
                ftnConvertHexStringToInteger = ftnConvertHexStringToInteger * 16
                intCharValue = ftnConvertHexCharToInteger(Mid(strHex, i, 1))
                ftnConvertHexStringToInteger += ftnConvertHexCharToInteger(Mid(strHex, i, 1))
            Next
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnConvertHexStringToInteger - " & ex.ToString)
        End Try
    End Function

    Function ftnConvertHexCharToInteger(ByRef strHex As String) As Integer
        ' returns -1 if non valid Hex chracter, else 
        Try
            ftnConvertHexCharToInteger = Asc(Left(strHex, 1)) - 48
            If (ftnConvertHexCharToInteger < 0) Or (ftnConvertHexCharToInteger > 22) Or ((ftnConvertHexCharToInteger > 9) And (ftnConvertHexCharToInteger < 17)) Then
                ftnConvertHexCharToInteger = -1
            ElseIf (ftnConvertHexCharToInteger > 9) Then
                ftnConvertHexCharToInteger -= 7
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnConvertHexCharToInteger - " & ex.ToString)
        End Try
    End Function




    Private Function ftnGetHeroSonarRange() As Single
        'Return scan range in Inches relative to center of Bot
        ftnGetHeroSonarRange = 0
        Try
            If boolHeroSelected Then
                Dim strRange As String = ftnHero_Communicate("GS", ">", 1000)
                ftnGetHeroSonarRange = ftnConvertHexStringToInteger(strRange, 2)
                If (ftnGetHeroSonarRange = 255) Then ftnGetHeroSonarRange = 0
                If (ftnGetHeroSonarRange > 0) Then _
                    ftnGetHeroSonarRange = ftnInchesToMapUnits(ftnGetHeroSonarRange * 0.408 + 3) ' 
                If boolDebugHero Then frmDebug.subWriteLogs(Color.Green, " Range = " & ftnGetHeroSonarRange & " inches")
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnGetHeroSonarRange")
        End Try
    End Function

    '================================ Movement Commands   =========================================

    Public Function ftnHeroTurn(ByRef sngTurnAngle As Single) As String
        Try
            ftnHeroTurn = ""
            strTypeOfMove = ""
            Dim intDistance As Integer = Round(Abs(sngTurnAngle) * 0.1624)
            If (sngTurnAngle >= 0) Then
                ftnHero_Communicate("GMC3E800", ">", 80000)  'Full Left
                intDistance = intDistance * (1 - sngSteeringBiasOffsetCorrection / 8)
            Else
                ftnHero_Communicate("GMC3EC93", ">", 80000) 'Full Right
                intDistance = intDistance * (1 + sngSteeringBiasOffsetCorrection / 8)
            End If
            Dim strDistanceInHex As String = Hex(intDistance)
            If (Len(strDistanceInHex) < 2) Then strDistanceInHex = "0" & strDistanceInHex
            strTypeOfMove = "Turn"
            ftnHeroTurn = ftnHero_Communicate("GMD308" & strDistanceInHex, ">", 10000 + intDistance * 1000) '
        Catch ex As Exception
            ftnHeroTurn = " COM ERROR"
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnHeroTurn")
        End Try
    End Function


    Public Function ftnHeroFwd(ByVal sngFwdDistance As Single, ByVal strParsedText As String()) As String
        Try
            ftnHeroFwd = ""
            sngFwdDistance = ftnMapUnitsToInches(sngFwdDistance)
            sngFwdDistance = sngFwdDistance * ftnFwdCorrectionFactor(sngFwdDistance)

            Dim strDistanceInHex As String = Hex(Round(Abs(sngFwdDistance) * 1.07))
            If (Len(strDistanceInHex) < 2) Then strDistanceInHex = "0" & strDistanceInHex
            Dim intPresentSteeringPosition As Short = Convert.ToInt32(ftnGetHeroPosition(6), 16) ' 
            strTypeOfMove = ""
            Dim intCorrectedCenter As Short = &H49 + sngSteeringBiasOffsetCorrection
            Dim strCorrectedCenter As String = Hex(intCorrectedCenter)
            If (intPresentSteeringPosition < intCorrectedCenter) Then
                ftnHero_Communicate("GMC3EC" & strCorrectedCenter, ">", 40000) ' Steer straight ahead
            ElseIf (intPresentSteeringPosition > intCorrectedCenter) Then
                ftnHero_Communicate("GMC3E8" & strCorrectedCenter, ">", 40000) ' Steer straight ahead
            End If
            strTypeOfMove = "Fwd"
            If boolEnableLocalizationWhileMoving Then subStartImageRecognition()
            timeMoveCommandSent = Now()
            boolBotIsMoving = True
            If (sngFwdDistance >= 0) Then
                ftnHeroFwd = ftnHero_Communicate("GMD308" & strDistanceInHex, ">", 10000 + sngFwdDistance * 1000) ' FWD 
            Else
                ftnHeroFwd = ftnHero_Communicate("GMD30C" & strDistanceInHex, ">", 10000 + Abs(sngFwdDistance) * 1000) ' BACK
            End If
            subWait(100)
            boolBotIsMoving = False
            If boolDebugHero Then ftnGetHeroPosition(8)
        Catch ex As Exception
            ftnHeroFwd = " COM ERROR"
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnHeroFwd")
        End Try
    End Function

    Function ftnFwdCorrectionFactor(ByRef sngDistance As Single) As Single
        ' Heathkit Heros are notoriously bad at driving the correct distance
        ' This is a crude attempt to correct the distance commanded in order to drive the distance desired
        Try
            ftnFwdCorrectionFactor = sngDistance / _
                                   (sngDistanceCorrectionCoeff_x2 * sngDistance ^ 2 _
                                   + sngDistanceCorrectionCoeff_x1 * sngDistance _
                                   + sngDistanceCorrectionCoeff_x0)
        Catch ex As Exception
            ftnFwdCorrectionFactor = 1
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnFwdCorrectionFactor")
        End Try
    End Function

    Public Function ftnHeroStop() As String
        Try
            ftnHeroStop = ftnHero_Communicate("GH", ">", 1000)
        Catch ex As Exception
            ftnHeroStop = " COM ERROR"
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnHeroStop")
        End Try
    End Function

    Public Function ftnGetHeroPosition(ByVal intMotorNumber As Short) As String
        'Return motor position (in Hex) for Hero
        ftnGetHeroPosition = "ERROR :"
        Try
            ftnGetHeroPosition = ftnHero_Communicate("GP" & intMotorNumber.ToString, ">", 1000)
            If boolDebugHero Then frmDebug.subWriteLogs(Color.Red, " Position for Motor " & intMotorNumber & " = " & ftnGetHeroPosition)
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnGetHeroPosition")
        End Try
    End Function



End Module
