﻿

' This module returns a bot pose based upon Range Scan Data.
' It is used for devices which can return a 2D range scan such as a Kinect or Laser Range scanner
Imports System
Imports System.Net
Imports System.Net.Sockets
Imports System.Text
Imports System.Threading
Imports Microsoft.VisualBasic
Imports System.EventArgs
Imports System.Math
'Imports Microsoft.Kinect

Module modLocalize2Dscan
    Dim bool2DScanLocalizationInProgress As Boolean = False ' dont want conflict 
    Dim intMaxRange As Short = ftnInchesToMapUnits(intScanVectorsMaxRangeInches)

    Public Function ftnLocalizeFrom2DScan_Top(ByRef tbpEstimatedPose As typeBotPose, ByVal tbpReferencePose As typeBotPose, ByVal boolBotIsMoving As Boolean, _
                                              ByRef boolGood2DLocalization As Boolean, ByRef sngRMS_Error As Single) As typeBotPose
        Try
            ftnLocalizeFrom2DScan_Top = tbpEstimatedPose
            boolGood2DLocalization = False
            If (ftnScannerEnabled() Or (boolRangersAreEnabled And Not boolBotIsMoving)) And Not boolStop And Not boolSimulatedMovement Then
                Dim tbpRawScanPose As typeBotPose
                sngRMS_Error = sngOneDotFiveSigma
                If boolBotIsMoving Or ftnIAmLost(tbpEstimatedPose) Then
                    intNumberOfBeams = Max(16, intNumberOfBeamsDefault \ 2) ' reduce number of vectors while moving or lost for faster localization
                Else
                    intNumberOfBeams = Max(16, intNumberOfBeamsDefault)
                End If
                intMaxRange = ftnInchesToMapUnits(intScanVectorsMaxRangeInches)
                boolAbortLocalization = False
                Dim strStatusDrivingRoot As String = strStatusDriving
                Dim boolGoodPreVal2DLocalization As Boolean = False
                If Not boolBotIsMoving Then subDoEventsIfMainThread()
                If bool2DScanLocalizationInProgress Then
                    Dim timeOut As DateTime = Now().AddMilliseconds(12000)
                    While bool2DScanLocalizationInProgress And Not (boolStop And Not (boolCollisionDetected Or boolCollisionPredicted)) ' wait for previous scan to finish
                        boolAbortLocalization = True
                        subDoEventsIfMainThread()
                        Thread.Sleep(25)
                        If Now() >= timeOut Then
                            ftnLocalizeFrom2DScan_Top = ftnLocalizeUsingDeadReckoning(ftnLocalizeFrom2DScan_Top, intScannerDelay, timeMoveCommandSent, tbpStartingPose, tbpTargetPoseGlobal, strTypeOfMove, True)
                            frmDebug.subWriteLogs(Color.Green, "  ftnLocalizeFrom2DScan_Top timeOut ")
                            Exit Function
                        End If
                    End While
                End If
                boolAbortLocalization = False
                bool2DScanLocalizationInProgress = True
                tbpRawScanPose = ftn2dLocalize(tbpEstimatedPose, boolGoodPreVal2DLocalization, sngRMS_Error)
                If Not boolGoodPreVal2DLocalization Then
                    ftnLocalizeFrom2DScan_Top = tbpEstimatedPose
                Else
                    If (sngRMS_Error < sngGoodMatch) Then
                        ftnLocalizeFrom2DScan_Top = tbpRawScanPose
                    ElseIf (sngRMS_Error < sngWeakMatch) Then
                        ftnLocalizeFrom2DScan_Top = ftnFindAvgPose(tbpEstimatedPose, tbpRawScanPose, 0)
                    End If
                    ftnLocalizeFrom2DScan_Top = ftnMinPoseUncertainty(ftnLocalizeFrom2DScan_Top)
                    If boolBotIsMoving Then tbpReferencePose = tbpEstimatedPose
                    ftnLocalizeFrom2DScan_Top = ftnValidatePose(ftnLocalizeFrom2DScan_Top, tbpReferencePose, boolGood2DLocalization)
                End If
                If Not boolBotIsMoving And Not boolEnableRBPF Then ' update graphics and log file with results
                    strStatusDriving = strStatusDrivingRoot & " 2D=" & FormatNumber(sngRMS_Error, 2) & "; "
                    GotABotClass.subRedraw(ftnLocalizeFrom2DScan_Top)
                    frmDebug.subWritePose(ftnElapsedTime(timeMoveCommandSent) & "       Pose Derived From 2D Scan Data  ", tbpRawScanPose)
                    If boolGood2DLocalization Then
                        subDrawScanVectors(tbpRawScanPose, True)
                        frmDebug.subWriteLogs(Color.Green, "                         Scan Coefficient = " & FormatNumber(sngRMS_Error, 2) & "   (2D localization success)")
                    Else
                        frmDebug.subWriteLogs(Color.Red, "                           Scan Coefficient = " & FormatNumber(sngRMS_Error, 2) & "   (2D localization failed)")
                        If boolDebugLocalize2D Then
                            frmDebug.subWritePose("tbpReferencePose = ", tbpReferencePose)
                            frmDebug.subWritePose("ftnLocalizeFrom2DScan_Top = ", ftnLocalizeFrom2DScan_Top)
                            If Not ftnLocationsEqual(ftnCheckMapArrayLimits(ftnLocalizeFrom2DScan_Top.Location), ftnLocalizeFrom2DScan_Top.Location) Then
                                frmDebug.subWriteLogs(Color.Red, " ftnValidatePose:ftnLocationsEqual = false")
                            ElseIf ftnInWall(ftnLocalizeFrom2DScan_Top.Location) Then
                                frmDebug.subWriteLogs(Color.Red, " ftnValidatePose:ftnInWall = false")
                            ElseIf (Not ftnLocationWithinPose(ftnLocalizeFrom2DScan_Top.Location, tbpReferencePose)) Then
                                frmDebug.subWriteLogs(Color.Red, " ftnValidatePose:ftnLocationWithinPose = false")
                            ElseIf Not ftnAnglesWithinUncertainty(ftnLocalizeFrom2DScan_Top.Orientation, tbpReferencePose.Orientation, tbpReferencePose.OrientationUncertainty) Then
                                frmDebug.subWriteLogs(Color.Red, " ftnValidatePose:ftnAnglesWithinUncertainty = false")
                            End If
                        End If
                    End If
                End If
            End If
            If boolDebugHero Or boolDebugLocalize2D Then
                Dim trvVector As typeRangeVector
                Dim tmcVectorOrigin As typeMapCoordinate = ftnLocationFromEstimatedPose(ftnLocalizeFrom2DScan_Top, tmcScanVectorOffset)
                Dim intDistanceFromWallErrorSquared As Short
                Dim boolEndPointInGreySpace As Boolean
                For intBeamNumber As Short = 0 To constMaxScannerBeams - 1 Step constMaxScannerBeams \ intNumberOfBeams
                    trvVector.sngRange = trvScanVectors(intBeamNumber).sngRange
                    If (trvVector.sngRange > 0) And (trvVector.sngRange < intMaxRange) Then  ' 
                        trvVector.sngAngle = ftnConvertTo360(trvScanVectors(intBeamNumber).sngAngle + ftnLocalizeFrom2DScan_Top.Orientation)
                        intDistanceFromWallErrorSquared = ftnDistanceFromWallErrorSquared(tmcVectorOrigin, trvVector, False, boolEndPointInGreySpace)
                        frmDebug.subWriteLogs(Color.Red, " Beam " & intBeamNumber & ":  Range = " & trvVector.sngRange & ":  Error = " & intDistanceFromWallErrorSquared)
                    End If
                Next intBeamNumber
            End If
            bool2DScanLocalizationInProgress = False
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnLocalizeFrom2DScan_Top")
        End Try
    End Function




    Dim boolftn2dLocalizeInProgress As Boolean = False
    Function ftn2dLocalize(ByRef tbpEstimatedPose As typeBotPose, _
                           ByRef boolGood2DLocalization As Boolean, ByRef sngRMS_Error As Single) As typeBotPose
        ftn2dLocalize = tbpEstimatedPose
        Try
            sngRMS_Error = sngOneDotFiveSigma
            boolGood2DLocalization = False
            If Not boolftn2dLocalizeInProgress Then
                boolftn2dLocalizeInProgress = True
                If Not boolBotIsMoving Then frmRangeFinder.subGetRangerReadings()
                If ftnScannerEnabled() Or (boolRangersAreEnabled And Not boolBotIsMoving) Then
                    If ftnGetScanVectors(tbpEstimatedPose, True) Then
                        If ftnNonFlatWall() Then ftn2dLocalize = ftnPoseFrom2DScan(tbpEstimatedPose, sngRMS_Error, False)
                        If Not boolHeroSelected And ((ftnDistanceBetweenLocations(tbpEstimatedPose.Location, ftn2dLocalize.Location) > sngTenInches) Or _
                               (Abs(ftnAngleDifference(tbpEstimatedPose.Orientation, ftn2dLocalize.Orientation)) > 12)) Then
                            boolGood2DLocalization = (sngRMS_Error <= sngGoodMatch)
                        Else
                            boolGood2DLocalization = (sngRMS_Error < sngWeakMatch)
                        End If
                    End If
                End If
            End If
            boolftn2dLocalizeInProgress = False
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftn2dLocalize :" & ex.ToString)
        End Try
    End Function




    Public Function ftnPoseFrom2DScan(ByVal tbpEstimatedPose As typeBotPose, ByRef sngRMS_Error As Single, ByRef boolEndPointModel As Boolean) As typeBotPose
        Try
            sngRMS_Error = sngOneDotFiveSigma
            ftnPoseFrom2DScan = tbpEstimatedPose
            If (tbpEstimatedPose.OrientationUncertainty > 25) Then
                ftnPoseFrom2DScan = ftnPoseWithLowestRangeError(ftnPoseFrom2DScan, sngRMS_Error, sngSixInches / sngGridSize, 6, boolEndPointModel)
                ftnPoseFrom2DScan.LocationUncertainty.X = sngSixInches / sngGridSize
                ftnPoseFrom2DScan.LocationUncertainty.Y = ftnPoseFrom2DScan.LocationUncertainty.X
                ftnPoseFrom2DScan.OrientationUncertainty = 8
            End If
            If (tbpEstimatedPose.OrientationUncertainty > 4) Then
                ftnPoseFrom2DScan = ftnPoseWithLowestRangeError(ftnPoseFrom2DScan, sngRMS_Error, sngFourInchesGrid, 3, boolEndPointModel)
                ftnPoseFrom2DScan.LocationUncertainty.X = sngThreeInchesGrid
                ftnPoseFrom2DScan.LocationUncertainty.Y = sngThreeInchesGrid
                ftnPoseFrom2DScan.OrientationUncertainty = 3
            End If
            If boolSlam Then
                If (sngRMS_Error < sngOneDotFiveSigma) Then ftnPoseFrom2DScan = ftnPoseWithLowestRangeError(ftnPoseFrom2DScan, sngRMS_Error, sngTwoInchesGrid, 2, boolEndPointModel)
            Else
                If (sngRMS_Error < sngOneDotFiveSigma) Then ftnPoseFrom2DScan = ftnPoseWithLowestRangeError(ftnPoseFrom2DScan, sngRMS_Error, sngOneInchGrid, 1, boolEndPointModel)
            End If
        Catch ex As Exception
            ftnPoseFrom2DScan = tbpEstimatedPose
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnPoseFrom2DScan")
        End Try
    End Function


    Dim boolAbortLocalization As Boolean = False
    Function ftnPoseWithLowestRangeError(ByRef tbpEstimatedPose As typeBotPose, ByRef sngRMS_ErrorBest As Single, _
                                                 ByVal intLocationStep As Short, ByVal intOrientationStep As Short, ByRef boolEndPointModel As Boolean) As typeBotPose
        Try ' Find the pose within the bot's uncertainty region that most closely resembles the range data
            ftnPoseWithLowestRangeError = tbpEstimatedPose
            sngRMS_ErrorBest = sngOneDotFiveSigma
            If boolStop And Not (boolCollisionDetected Or boolCollisionPredicted) Or boolAbortLocalization Then Exit Function
            Dim intOrientationStart, intOrientationEnd As Short
            Dim tbpPossiblePose As typeBotPose = tbpEstimatedPose
            Dim intStartX, intEndX, intStartY, intEndY As Short
            Dim sngRMS_Error, sngRMS_ErrorEndPoint, sngRMS_ErrorEndPointBest As Single
            intLocationStep = Min(intLocationStep, 1)
            intStartX = tbpEstimatedPose.Location.X - tbpEstimatedPose.LocationUncertainty.X
            intEndX = tbpEstimatedPose.Location.X + tbpEstimatedPose.LocationUncertainty.X
            intStartY = tbpEstimatedPose.Location.Y - tbpEstimatedPose.LocationUncertainty.Y
            intEndY = tbpEstimatedPose.Location.Y + tbpEstimatedPose.LocationUncertainty.Y
            subCheckMapArrayLimits(intStartX, intEndX, intStartY, intEndY)
            subFindStartEndAngles(tbpEstimatedPose, intOrientationStart, intOrientationEnd)
            sngRMS_ErrorBest = ftnRMS_RangeError(tbpEstimatedPose, boolEndPointModel) ' Find  error at estimated pose
            For tbpPossiblePose.Location.X = intStartX To intEndX Step intLocationStep ' try to find a lower error pose 
                For tbpPossiblePose.Location.Y = intStartY To intEndY Step intLocationStep
                    If Not ftnInWall(tbpPossiblePose.Location) Then
                        For intOrientation As Short = intOrientationStart To intOrientationEnd Step intOrientationStep
                            tbpPossiblePose.Orientation = ftnConvertTo360(intOrientation)
                            sngRMS_ErrorEndPoint = ftnRMS_RangeError(tbpPossiblePose, True) ' Find End Point Model error at this pose
                            If (sngRMS_ErrorEndPoint < sngRMS_ErrorBest) Then ' check for intervening walls
                                If boolEndPointModel Then  ' Dont perform Ray Casting for every possible pose  - its too slow.
                                    sngRMS_ErrorBest = sngRMS_ErrorEndPoint
                                    ftnPoseWithLowestRangeError = tbpPossiblePose
                                Else
                                    sngRMS_Error = ftnRMS_RangeError(tbpPossiblePose, False) ' Find Ray Cast Model error at this pose
                                    If (sngRMS_Error < sngRMS_ErrorBest) Then ' 
                                        ftnPoseWithLowestRangeError = tbpPossiblePose
                                        sngRMS_ErrorEndPointBest = sngRMS_ErrorEndPoint
                                        sngRMS_ErrorBest = sngRMS_Error
                                    End If
                                End If
                            End If
                        Next intOrientation
                    End If
                Next tbpPossiblePose.Location.Y
                subDoEventsIfMainThread()
                If boolStop And Not (boolCollisionDetected Or boolCollisionPredicted) Or boolAbortLocalization Then
                    ftnPoseWithLowestRangeError = tbpEstimatedPose
                    sngRMS_ErrorBest = sngOneDotFiveSigma
                    frmDebug.subWriteLogs(Color.Green, "  >>>>>>>>>>>>>>>>>> WARNING:  ftnPoseWithLowestRangeError aborted")
                    Exit Function
                End If
            Next tbpPossiblePose.Location.X
            If Not boolEndPointModel Then
                sngRMS_ErrorBest = ftnRMS_RangeError(ftnPoseWithLowestRangeError, False) ' Find Ray Cast Model error at this pose
            Else
                sngRMS_ErrorBest = sngRMS_ErrorBest * 1.3 ' 1.3 is a "magic number" used to convert End Point errors to Ray Cast
            End If
            sngRMS_ErrorBest = ftnMapUnitsToInches(sngRMS_ErrorBest)
        Catch ex As Exception
            ftnPoseWithLowestRangeError = tbpEstimatedPose
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnPoseWithLowestRangeError")
        End Try
    End Function

    Function ftnRMS_RangeError(ByRef tbpPossiblePose As typeBotPose, ByRef boolEndPointModel As Boolean) As Single
        ' boolEndPointModel = Don't check for intervening walls
        Try
            ftnRMS_RangeError = ftnScanToMapRMS_Error(tbpPossiblePose, boolEndPointModel)
            If Not boolBotIsMoving And boolRangersAreEnabled Then
                Dim intRangerValidBeams, intNumberOfRangers As Short
                Dim trvRangerBeams(intHighestEnabledRangerNumber) As typeRangeVector  ' Range vector - relative to Bot center
                trvRangerBeams = ftnRangerBeamsRelativeBotCenter(tbpPossiblePose)
                Dim sngRMS_RangerError As Single = ftnRangerToMapRMS_Error(tbpPossiblePose, intRangerValidBeams, intNumberOfRangers, boolEndPointModel, trvRangerBeams) ' 
                If (ftnRMS_RangeError < sngOneDotFiveSigma) And (sngRMS_RangerError < sngOneDotFiveSigma) Then
                    ftnRMS_RangeError = Sqrt((ftnRMS_RangeError * intNumberOfBeams) ^ 2 + (sngRMS_RangerError * intNumberOfRangers) ^ 2) / (intNumberOfBeams + intNumberOfRangers)
                ElseIf (intRangerValidBeams >= 12) Then
                    ftnRMS_RangeError = sngRMS_RangerError
                End If
            End If
        Catch ex As Exception
            ftnRMS_RangeError = sngOneDotFiveSigma
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnRMS_RangeError")
        End Try
    End Function

    Function ftnRangerBeamsRelativeBotCenter(ByRef tbpEstimatedPose As typeBotPose) As typeRangeVector() ' Find Range vectors - relative to Bot center
        Dim trvRangerVector(intHighestEnabledRangerNumber) As typeRangeVector
        Try
            If boolRangersAreEnabled Then
                For intRangerNumber As Short = 1 To intHighestEnabledRangerNumber ' 
                    If trfRangeFinders(intRangerNumber).boolEnable Then _
                        trvRangerVector(intRangerNumber) = frmRangeFinder.ftnRangeFinderVector(tbpEstimatedPose, intRangerNumber, trfRangeFinders(intRangerNumber).sngRange, trfRangeFinders(intRangerNumber).intOrientation)
                Next intRangerNumber
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnRangerBeamsRelativeBotCenter")
        End Try
        Return trvRangerVector
    End Function


    Function ftnScanToMapRMS_Error(ByRef tbpPossiblePose As typeBotPose, ByRef boolEndPointModel As Boolean) As Single
        ' Find the correlation of a particular pose to the range data 
        ' boolEndPointModel = Don't check for intervening walls
        Dim intDistanceFromWallErrorSquared As Short
        Dim intValidBeamCounter As Short = 0
        Dim intDistanceFromWallErrorSquaredSum As Short = 0
        Dim boolEndPointInGreySpace As Boolean = False
        Try
            If ftnScannerEnabled() Then
                ftnScanToMapRMS_Error = 0
                Dim trvVector As typeRangeVector
                Dim tmcVectorOrigin As typeMapCoordinate = ftnLocationFromEstimatedPose(tbpPossiblePose, tmcScanVectorOffset)
                For intBeamNumber As Short = 0 To constMaxScannerBeams - 1 Step constMaxScannerBeams \ intNumberOfBeams
                    trvVector.sngRange = trvScanVectors(intBeamNumber).sngRange
                    If (trvVector.sngRange > 0) And (trvVector.sngRange < intMaxRange) Then  ' 
                        trvVector.sngAngle = ftnConvertTo360(trvScanVectors(intBeamNumber).sngAngle + tbpPossiblePose.Orientation)
                        intDistanceFromWallErrorSquared = ftnDistanceFromWallErrorSquared(tmcVectorOrigin, trvVector, boolEndPointModel, boolEndPointInGreySpace)
                        If (intDistanceFromWallErrorSquared < sngOneSigmaSquared) Then
                            intValidBeamCounter += 3
                        ElseIf Not boolSlam Then
                            If boolEndPointInGreySpace Then intValidBeamCounter += 1 ' Grey zone = cluttered environment / may have false readings
                        End If
                        If boolHeroSelected And Not boolKinectConnected And Not boolURGConnected Then intDistanceFromWallErrorSquared = intDistanceFromWallErrorSquared / 2
                        intDistanceFromWallErrorSquaredSum += intDistanceFromWallErrorSquared
                    End If
                Next intBeamNumber
                intValidBeamCounter = intValidBeamCounter \ 3
                If (intValidBeamCounter > intNumberOfBeams * 0.6) Or boolHeroSelected And (intValidBeamCounter > intNumberOfBeams * 0.4) Then
                    ftnScanToMapRMS_Error = Sqrt(intDistanceFromWallErrorSquaredSum / intValidBeamCounter)
                    If boolURGConnected Then
                        ftnScanToMapRMS_Error *= 0.7  ' wider FOV then Kinects yields poorer fit, therefore higher error??
                    ElseIf boolHeroSelected Then
                        ftnScanToMapRMS_Error *= 0.7
                    End If
                    ftnScanToMapRMS_Error = Min(ftnScanToMapRMS_Error, sngOneDotFiveSigma)
                    ftnScanToMapRMS_Error = Max(ftnScanToMapRMS_Error, 0.25)
                Else
                    ftnScanToMapRMS_Error = sngOneDotFiveSigma
                End If
            Else
                ftnScanToMapRMS_Error = sngOneDotFiveSigma
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Green, " EXECEPTION in ftnScanToMapRMS_Error")
            ftnScanToMapRMS_Error = sngOneDotFiveSigma
        End Try
    End Function

    Function ftnRangerToMapRMS_Error(ByVal tbpEstimatedPose As typeBotPose, ByRef intValidBeamCounter As Short, ByRef intNumberOfRangers As Short, _
                                       ByVal boolEndPointModel As Boolean, ByVal trvRangerBeams() As typeRangeVector) As Single
        ' Find the correlation of a particular pose to the range data (higher =  better)
        ' boolEndPointModel = Don't check for intervening walls
        Try
            Dim boolEndPointInGreySpace As Boolean = False
            intValidBeamCounter = 0
            intNumberOfRangers = 0
            If boolRangersAreEnabled Then
                ftnRangerToMapRMS_Error = 0
                Dim intDistanceFromWallErrorSquared As Short
                Dim trvVector As typeRangeVector
                For intRangerNumber As Short = 1 To intHighestEnabledRangerNumber ' 
                    If trfRangeFinders(intRangerNumber).boolEnable Then
                        intNumberOfRangers += 1
                        trvVector.sngRange = trvRangerBeams(intRangerNumber).sngRange
                        trvVector.sngAngle = ftnConvertTo360(Convert.ToInt16(trvRangerBeams(intRangerNumber).sngAngle) + tbpEstimatedPose.Orientation)
                        intDistanceFromWallErrorSquared = Abs(ftnDistanceFromWallErrorSquared(tbpEstimatedPose.Location, trvVector, boolEndPointModel, boolEndPointInGreySpace))
                        ftnRangerToMapRMS_Error += intDistanceFromWallErrorSquared
                        If (intDistanceFromWallErrorSquared >= intOneDotFiveSigmaSquared) Then
                            If Not boolSlam And boolEndPointInGreySpace Then intValidBeamCounter += 2 ' Grey zone = cluttered environment / may have false echos
                        Else
                            intValidBeamCounter += 3
                        End If
                    End If
                Next intRangerNumber
                intValidBeamCounter = intValidBeamCounter \ 3
                If (intValidBeamCounter >= (intNumberOfRangers \ 2)) Then
                    ftnRangerToMapRMS_Error = Min(Sqrt(ftnRangerToMapRMS_Error) / intNumberOfRangers, sngOneDotFiveSigma)
                Else
                    ftnRangerToMapRMS_Error = sngOneDotFiveSigma
                End If
            Else
                ftnRangerToMapRMS_Error = sngOneDotFiveSigma
            End If
        Catch ex As Exception
            ftnRangerToMapRMS_Error = sngOneDotFiveSigma
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnRangerToMapRMS_Error")
        End Try
    End Function

    Function ftnPoseUncertaintyFrom2DScan(ByRef tbpEstimatedPose As typeBotPose, _
                                          ByRef sngRMS_RangeError As Single) As typeBotPose
        ' Reduce pose uncertainty if we have a good match        
        ftnPoseUncertaintyFrom2DScan = tbpEstimatedPose
        If (sngRMS_RangeError <= sngExcellentMatch) Then
            ftnPoseUncertaintyFrom2DScan.LocationUncertainty.X = Min(ftnPoseUncertaintyFrom2DScan.LocationUncertainty.X, sngTenInchesGrid)
            ftnPoseUncertaintyFrom2DScan.LocationUncertainty.Y = Min(ftnPoseUncertaintyFrom2DScan.LocationUncertainty.Y, ftnPoseUncertaintyFrom2DScan.LocationUncertainty.X)
            ftnPoseUncertaintyFrom2DScan.OrientationUncertainty = Min(ftnPoseUncertaintyFrom2DScan.OrientationUncertainty, 10)
        ElseIf (sngRMS_RangeError <= sngGoodMatch) Then
            ftnPoseUncertaintyFrom2DScan.LocationUncertainty.X = Min(ftnPoseUncertaintyFrom2DScan.LocationUncertainty.X, sngTwelveInches / sngGridSize)
            ftnPoseUncertaintyFrom2DScan.LocationUncertainty.Y = Min(ftnPoseUncertaintyFrom2DScan.LocationUncertainty.Y, ftnPoseUncertaintyFrom2DScan.LocationUncertainty.X)
            ftnPoseUncertaintyFrom2DScan.OrientationUncertainty = Min(ftnPoseUncertaintyFrom2DScan.OrientationUncertainty, 12)
        ElseIf (sngRMS_RangeError < sngWeakMatch) Then
            ftnPoseUncertaintyFrom2DScan.LocationUncertainty.X = Min(ftnPoseUncertaintyFrom2DScan.LocationUncertainty.X, sngFifteenInches / sngGridSize)
            ftnPoseUncertaintyFrom2DScan.LocationUncertainty.Y = Min(ftnPoseUncertaintyFrom2DScan.LocationUncertainty.Y, ftnPoseUncertaintyFrom2DScan.LocationUncertainty.X)
            ftnPoseUncertaintyFrom2DScan.OrientationUncertainty = Min(ftnPoseUncertaintyFrom2DScan.OrientationUncertainty, 16)
        End If
        ftnPoseUncertaintyFrom2DScan = ftnMinPoseUncertainty(ftnPoseUncertaintyFrom2DScan)
    End Function



    '--------------------------------------  Non Flat Wall  ----------------------------------------------------------------------

    Public Function ftnNonFlatWall() As Boolean
        ' long flat walls which fill the Scanner's FOV can cause localization errors
        ftnNonFlatWall = False
        Try
            Dim sngScannerRange(constMaxScannerBeams) As Single
            Dim sngScannerSlope(constMaxScannerBeams) As Single
            Dim intBeamInterval As Short = constMaxScannerBeams \ intNumberOfBeams
            Dim intTwoSamplesPerVector As Short = intBeamInterval + intBeamInterval
            For intBeamNumber As Short = 0 To constMaxScannerBeams - 1 Step intBeamInterval
                sngScannerRange(intBeamNumber) = trvScanVectors(intBeamNumber).sngRange
            Next intBeamNumber
            If (sngScannerRange(intTwoSamplesPerVector) = 0) Then sngScannerRange(intTwoSamplesPerVector) = sngScannerRange(3 * intBeamInterval)
            If (sngScannerRange(intBeamInterval) = 0) Then sngScannerRange(intBeamInterval) = sngScannerRange(intTwoSamplesPerVector)
            If (sngScannerRange(0) = 0) Then sngScannerRange(0) = sngScannerRange(intBeamInterval)
            sngScannerSlope(intBeamInterval) = sngScannerRange(intBeamInterval) - sngScannerRange(0)
            Dim intI_MinusSamplesPerVector As Short
            For I As Short = intTwoSamplesPerVector To constMaxScannerBeams - 1 Step intBeamInterval
                intI_MinusSamplesPerVector = I - intBeamInterval
                If (sngScannerRange(I) = 0) Then sngScannerRange(I) = sngScannerRange(intI_MinusSamplesPerVector) ' get rid of range = 0 entries
                sngScannerSlope(I) = sngScannerRange(I) - sngScannerRange(intI_MinusSamplesPerVector) ' calculate slopes
                If (sngScannerRange(I) <> 0) And (sngScannerRange(intI_MinusSamplesPerVector) <> 0) Then
                    If (Abs(sngScannerRange(I) - sngScannerRange(intI_MinusSamplesPerVector)) > sngThreeInches) Then ' Exit if Spike (Discontinuity)
                        ftnNonFlatWall = True
                        Exit Function
                    End If
                End If
            Next I
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnNonFlatWall")
        End Try
    End Function




End Module
