﻿
Imports System.Math

Module modLocalizeDeadReckoning


    Public Function ftnLocalizeUsingDeadReckoning(ByRef tbpEstimatedPose As typeBotPose, ByRef sngSensorDelay_seconds As Single, ByRef timeMoveCommandSent As DateTime, _
                                                  ByRef tbpStartingPose As typeBotPose, ByRef tbpTargetPose As typeBotPose, ByRef strTypeOfMove As String, ByRef boolCalculateUncertainties As Boolean) As typeBotPose
        ' sngSensorDelay_seconds is used to calculate the pose at the time sensor readings (Kinect, IR, camera, whatever) were measured
        Try
            Dim sngDistanceFromStart, sngAngleTraveled, sngElapsedTime As Single
            sngElapsedTime = ftnElapsedTime(timeMoveCommandSent) - sngSensorDelay_seconds  ' in Seconds
            If (String.CompareOrdinal(strTypeOfMove, "Turn") = 0) Then
                sngAngleTraveled = ftnElapsedAngle(tbpStartingPose, tbpTargetPose, sngElapsedTime)
                ftnLocalizeUsingDeadReckoning = ftnTurnDR(tbpStartingPose, sngAngleTraveled, boolCalculateUncertainties)
            ElseIf (String.CompareOrdinal(strTypeOfMove, "Fwd") = 0) Then
                sngDistanceFromStart = ftnDistanceFromStart(tbpStartingPose, tbpTargetPose, sngElapsedTime, strTypeOfMove)
                ftnLocalizeUsingDeadReckoning = ftnFwdDR(tbpStartingPose, sngDistanceFromStart, boolCalculateUncertainties)
            ElseIf (String.CompareOrdinal(strTypeOfMove, "XY") = 0) Then
                sngDistanceFromStart = ftnDistanceFromStart(tbpStartingPose, tbpTargetPose, sngElapsedTime, strTypeOfMove)
                ftnLocalizeUsingDeadReckoning = ftnXY_DR(tbpStartingPose, sngDistanceFromStart, tbpTargetPose, boolCalculateUncertainties)
            Else
                ftnLocalizeUsingDeadReckoning = tbpEstimatedPose
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnLocalizeUsingDeadReckoning")
        End Try
    End Function

    Public Function ftnFwdDR(ByRef tbpStartingPose As typeBotPose, ByRef sngDistanceFromStart As Single, ByRef boolCalculateUncertainties As Boolean) As typeBotPose
        Try
            ftnFwdDR = tbpStartingPose
            Dim sngOrientationRadians As Single = ftnDegreeToRadian(tbpStartingPose.Orientation)
            Dim sngDistanceFromStartGrid As Single = sngDistanceFromStart / sngGridSize
            Dim intDeltaX As Integer = sngDistanceFromStartGrid * System.Math.Cos(sngOrientationRadians)
            Dim intDeltaY As Integer = sngDistanceFromStartGrid * System.Math.Sin(-sngOrientationRadians)
            ftnFwdDR.Orientation = ftnConvertTo360(tbpStartingPose.Orientation)
            ftnFwdDR = tbpStartingPose
            ftnFwdDR.Location.X += intDeltaX
            ftnFwdDR.Location.Y += intDeltaY
            If boolCalculateUncertainties Then
                Dim sngUncertaintyFactor As Single = sngBotMovementAccuracy * sngTenInches
                ftnFwdDR.LocationUncertainty.X = tbpStartingPose.LocationUncertainty.X + Abs(intDeltaX / sngUncertaintyFactor)
                ftnFwdDR.LocationUncertainty.Y = tbpStartingPose.LocationUncertainty.Y + Abs(intDeltaY / sngUncertaintyFactor)
                ftnFwdDR.OrientationUncertainty = tbpStartingPose.OrientationUncertainty + Abs(ftnMapUnitsToInches(sngDistanceFromStart) / (sngBotMovementAccuracy * 20))
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnFwdDR")
        End Try
    End Function



    Public Function ftnXY_DR(ByRef tbpStartingPose As typeBotPose, ByRef sngDistanceFromStart As Single, ByRef tbpTargetPose As typeBotPose, ByRef boolCalculateUncertainties As Boolean) As typeBotPose
        Try
            ' MANY Assumptions - XY Motion = drive forward while turning towards target, then drive straight toward target
            '                    Assumptions:  Radius of turn = diameter of Bot;  Speed = Constant for turning and forward motion;  
            ' see http://rossum.sourceforge.net/papers/CalculationsForRobotics/CirclePath.htm  (but his equations do not agree with mine)
            '
            ' RIGHT TURN
            ' X(t) = x(start) - radius * sin(tbpStartingPose.orientation) + radius *sin(Anglular Velocity * t - tbpStartingPose.orientation)
            ' Y(t) = y(start) + radius * cos(tbpStartingPose.orientation) - radius *cos(-Anglular Velocity * t + tbpStartingPose.orientation)
            '
            'LEFT TURN
            ' X(t) = x(center) - radius *sin(Anglular Velocity * t + tbpStartingPose.orientation - 180)
            ' Y(t) = y(center) - radius *cos(Anglular Velocity * t + tbpStartingPose.orientation - 180)
            ftnXY_DR = tbpStartingPose

            Dim sngAngleOfTurn As Single = ftnTurnAngle(tbpStartingPose, tbpTargetPose)
            Dim boolTurnRight As Boolean = (sngAngleOfTurn < 0)
            sngAngleOfTurn = Abs(sngAngleOfTurn)
            Dim sngAngleOfTurnRadians As Single = ftnDegreeToRadian(sngAngleOfTurn)
            Dim sngXY_Radius As Single = sngBotDiameter '   ' BIG Assumption
            Dim sngXY_RadiusGrid As Single = sngXY_Radius / sngGridSize '
            Dim sngTurnDistance As Single = Abs(sngXY_Radius * sngAngleOfTurnRadians) ' NOTE: Simplified - no linear or angular velocities
            Dim sngTurnPercentage As Single = 1
            If (sngDistanceFromStart < sngTurnDistance) Then sngTurnPercentage = (sngDistanceFromStart / sngTurnDistance)
            Dim sngAngleOfTurnPercent As Single = sngAngleOfTurn * sngTurnPercentage
            Dim sngAngleOfTurnPercentRadians = (sngAngleOfTurnRadians * sngTurnPercentage)
            Dim StartOrientationRadians As Single = ftnConvertTo180(ftnDegreeToRadian(ftnXY_DR.Orientation))

            If boolTurnRight Then    ' Find the pose on the Turning portion of the move
                ftnXY_DR.Location.X = tbpStartingPose.Location.X + sngXY_RadiusGrid * (Sin(StartOrientationRadians) + Sin(sngAngleOfTurnPercentRadians - StartOrientationRadians))
                ftnXY_DR.Location.Y = tbpStartingPose.Location.Y + sngXY_RadiusGrid * (Cos(StartOrientationRadians) - Cos(-sngAngleOfTurnPercentRadians + StartOrientationRadians))
                ftnXY_DR.Orientation = ftnConvertTo360(tbpStartingPose.Orientation - Convert.ToInt16(sngAngleOfTurnPercent))
            Else
                Dim tmcCenterOfTurn As typeMapCoordinate
                tmcCenterOfTurn = ftnLocationFromEstimatedPose(tbpStartingPose, 90, sngXY_Radius)
                ftnXY_DR.Location.X = tmcCenterOfTurn.X - sngXY_RadiusGrid * Sin(StartOrientationRadians + sngAngleOfTurnPercentRadians - 3.14159)
                ftnXY_DR.Location.Y = tmcCenterOfTurn.Y - sngXY_RadiusGrid * Cos(StartOrientationRadians + sngAngleOfTurnPercentRadians - 3.14159)   '  
                ftnXY_DR.Orientation = ftnConvertTo360(tbpStartingPose.Orientation + Convert.ToInt16(sngAngleOfTurnPercent))
            End If

            If (sngTurnPercentage >= 1) Then  ' Find the pose on the Straight portion of the move
                ftnXY_DR.Orientation = ftnP1ToP2Angle(ftnXY_DR.Location, tbpTargetPose.Location)
                Dim sngDistanceFromCurveEndToTarget As Single = ftnDistanceBetweenLocations(ftnXY_DR.Location, tbpTargetPose.Location)
                ftnXY_DR.Location = ftnLocationFromEstimatedPose(ftnXY_DR, 0, Min((sngDistanceFromStart - sngTurnDistance), sngDistanceFromCurveEndToTarget))
            End If

            If boolCalculateUncertainties Then
                Dim intDeltaX As Integer = ftnXY_DR.Location.X - tbpStartingPose.Location.X
                Dim intDeltaY As Integer = ftnXY_DR.Location.Y - tbpStartingPose.Location.Y
                Dim sngUncertaintyFactor As Single = sngBotMovementAccuracy * sngTenInches
                ftnXY_DR.LocationUncertainty.X = tbpStartingPose.LocationUncertainty.X + Abs(intDeltaX / sngUncertaintyFactor)
                ftnXY_DR.LocationUncertainty.Y = tbpStartingPose.LocationUncertainty.Y + Abs(intDeltaY / sngUncertaintyFactor)
                ftnXY_DR.OrientationUncertainty = Max(16, tbpStartingPose.OrientationUncertainty + Min(Abs(sngAngleOfTurn) / (sngBotMovementAccuracy * 8), 8))
                If ftnLocationsEqual(ftnXY_DR.Location, tbpTargetPose.Location) Then ftnXY_DR.Orientation = tbpTargetPose.Orientation
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnXY_DR")
        End Try
    End Function

    Public Function ftnXY_TurnCompleted_DR(ByRef tbpStartingPose As typeBotPose, ByRef tbpTargetPose As typeBotPose) As Boolean
        Try
            ftnXY_TurnCompleted_DR = False
            Dim sngElapsedTime As Single = ftnElapsedTime(timeMoveCommandSent)
            Dim sngDistanceFromStart As Single = ftnDistanceFromStart(tbpStartingPose, tbpTargetPose, sngElapsedTime, "XY")
            Dim sngAngleOfTurn As Single = Abs(ftnTurnAngle(tbpStartingPose, tbpTargetPose))
            Dim sngAngleOfTurnRadians As Single = ftnDegreeToRadian(sngAngleOfTurn)
            Dim sngXY_Radius As Single = sngBotDiameter '   ' BIG Assumption
            Dim sngTurnDistance As Single = Abs(sngXY_Radius * sngAngleOfTurnRadians) ' NOTE: Simplified - no linear or angular velocities
            Dim sngTurnPercentage As Single = 1
            If (sngDistanceFromStart < sngTurnDistance) Then sngTurnPercentage = (sngDistanceFromStart / sngTurnDistance)

            ftnXY_TurnCompleted_DR = (sngTurnPercentage >= 1)
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnXY_TurnCompleted_DR")
        End Try
    End Function





    Function ftnTurnDR(ByRef tbpStartingPose As typeBotPose, ByRef intTurnAngle As Short, ByRef boolCalculateUncertainties As Boolean) As typeBotPose
        Try
            ftnTurnDR = tbpStartingPose
            ftnTurnDR.Orientation = ftnConvertTo360(tbpStartingPose.Orientation + intTurnAngle)
            If boolCalculateUncertainties Then
                If boolER1Selected And (sngWheelSeparation < 33) Then
                    ftnTurnDR.OrientationUncertainty += Abs(ftnConvertTo180(intTurnAngle) / (sngBotMovementAccuracy * 12))
                Else
                    ftnTurnDR.OrientationUncertainty += Abs(ftnConvertTo180(intTurnAngle) / (sngBotMovementAccuracy * 20))
                End If
                Dim sngLocationUncertaintyIncrease As Single = sngOneInchGrid
                ftnTurnDR.LocationUncertainty.X += sngLocationUncertaintyIncrease
                ftnTurnDR.LocationUncertainty.Y += sngLocationUncertaintyIncrease
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnTurnDR")
        End Try
    End Function

    Public Function ftnElapsedTime(ByRef timeMoveCommandSent As DateTime) As Single ' returns elapsed time since start of move in seconds
        Try
            Dim tsElapsedTimeSpan As TimeSpan = Now().Subtract(timeMoveCommandSent)
            ftnElapsedTime = (Convert.ToSingle(tsElapsedTimeSpan.TotalMilliseconds.ToString) - sngStartOfMoveDelay) * 0.001
            If (ftnElapsedTime < 0) Then ftnElapsedTime = 0
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnElapsedTime")
        End Try
    End Function


    Function ftnDistanceFromStart(ByRef tbpStartingPose As typeBotPose, ByRef tbpTargetPose As typeBotPose, ByRef sngElapsedTime As Single, ByRef strTypeOfMove As String) As Single
        ' d = vt + (1/2)at^2 
        ' v = a t  
        Try
            Dim sngDistanceCommanded As Single = ftnDistanceCommanded(tbpStartingPose, tbpTargetPose)
            If (String.CompareOrdinal(strTypeOfMove, "XY") = 0) Then sngDistanceCommanded = sngDistanceCommanded * 2 '/ Cos(ftnDegreeToRadian(sngAngleOfTurn)) ' 
            If (sngElapsedTime > sngBotLinearAccelerationTime) Then
                If (String.CompareOrdinal(strTypeOfMove, "XY") = 0) Then
                    ftnDistanceFromStart = sngBotLinearAccelerationDistanceXY + ftnCmToMapUnits((sngElapsedTime - sngBotLinearAccelerationTimeXY) * sngBotLinearVelocityXY)
                Else
                    ftnDistanceFromStart = sngBotLinearAccelerationDistance + ftnCmToMapUnits((sngElapsedTime - sngBotLinearAccelerationTime) * sngBotLinearVelocity)
                End If
            Else ' TODO Add deacceleration distance
                If (String.CompareOrdinal(strTypeOfMove, "XY") = 0) Then
                    ftnDistanceFromStart = ftnCmToMapUnits(sngBotLinearAcceleration * sngElapsedTime * sngElapsedTime * 0.5)
                Else
                    ftnDistanceFromStart = ftnCmToMapUnits(sngBotLinearAcceleration * sngElapsedTime * sngElapsedTime * 0.5)
                End If
            End If
            If (ftnDistanceFromStart < 0) Then ftnDistanceFromStart = 0
            If (ftnDistanceFromStart > Abs(sngDistanceCommanded)) Then ftnDistanceFromStart = Abs(sngDistanceCommanded)
            If (sngDistanceCommanded < 0) Then ftnDistanceFromStart = -ftnDistanceFromStart
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnDistanceFromStart")
        End Try
    End Function

    Function ftnElapsedAngle(ByRef tbpStartingPose As typeBotPose, ByRef tbpTargetPose As typeBotPose, ByRef sngElapsedTime As Single) As Single
        Try
            Dim sngTurnAngleCommanded As Single = ftnAngleDifference(tbpTargetPose.Orientation, tbpStartingPose.Orientation)
            If (sngElapsedTime > sngBotAngularAccelerationTime) Then
                ftnElapsedAngle = sngBotAngularAccelerationDistance + (sngElapsedTime - sngBotAngularAccelerationTime) * sngBotAngularVelocity
            Else
                ftnElapsedAngle = sngBotAngularAcceleration * sngElapsedTime * sngElapsedTime * 0.5
            End If
            If (ftnElapsedAngle < 0) Then ftnElapsedAngle = 0
            If (ftnElapsedAngle > Abs(sngTurnAngleCommanded)) Then ftnElapsedAngle = Abs(sngTurnAngleCommanded)
            If (sngTurnAngleCommanded < 0) Then ftnElapsedAngle = -ftnElapsedAngle
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnElapsedAngle")
        End Try
    End Function



End Module



