﻿Imports System.Math


Public Module modLocalizeOdometry
    ' This module is responsible for robot localization through odometry data.

    Public tmcStartingOdometry As typeMapCoordinate

    Public Function ftnLocalizeUsingOdometry(ByRef tbpEstimatedPose As typeBotPose, ByRef tbpStartingPose As typeBotPose) As typeBotPose
        ftnLocalizeUsingOdometry = tbpEstimatedPose
        Try  '
            If boolER1Selected And Not boolEnableBotSerialIF And (String.CompareOrdinal(strTypeOfMove, "Fwd") = 0) Then
                Dim boolGoodLocalization As Boolean = True
                Dim tmcCurrentOdometry As typeMapCoordinate = ftnGetOdometry(tbpEstimatedPose, tbpStartingPose)
                Dim trvOdometryVector As typeRangeVector
                trvOdometryVector.sngRange = (ftnDistanceBetweenLocations(tmcStartingOdometry, tmcCurrentOdometry))
                Dim sngDistanceCommanded As Single = ftnDistanceCommanded(tbpStartingPose, tbpTargetPoseGlobal)
                If (sngDistanceCommanded < 0) Then trvOdometryVector.sngRange = -trvOdometryVector.sngRange
                trvOdometryVector.sngAngle = tbpEstimatedPose.Orientation
                ftnLocalizeUsingOdometry.Location = ftnLocationFromLocation(tbpStartingPose.Location, trvOdometryVector.sngAngle, trvOdometryVector.sngRange)
                ftnLocalizeUsingOdometry.Orientation = trvOdometryVector.sngAngle
                If boolDebugOdometry Then frmDebug.subWritePose("       Raw Pose Derived From Odometry  ", ftnLocalizeUsingOdometry)
                ftnLocalizeUsingOdometry = ftnValidatePose(ftnLocalizeUsingOdometry, tbpEstimatedPose, boolGoodLocalization)
                tmcStartingOdometry = tmcCurrentOdometry
                If Not boolBotIsMoving Then _
                    frmDebug.subWritePose(ftnElapsedTime(timeMoveCommandSent) & "       Pose Derived From Odometry Data ", ftnLocalizeUsingOdometry)
            Else
                tmcStartingOdometry = ftnGetOdometry(tbpEstimatedPose, tbpStartingPose)
            End If
        Catch ex As Exception
            ftnLocalizeUsingOdometry = tbpEstimatedPose
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnLocalizeUsingOdometry " & ex.ToString)
        End Try
    End Function

    Function ftnGetOdometry(ByRef tbpEstimatedPose As typeBotPose, ByRef tbpStartingPose As typeBotPose) As typeMapCoordinate ' returns a relative map coordinate
        Try
            If boolER1Selected And Not boolEnableBotSerialIF Then
                ftnGetOdometry = ftnGetOdometry_ER1(tbpEstimatedPose, tbpStartingPose)
            Else ' use Dead Reckoning
                ftnGetOdometry.X = 9999
                ftnGetOdometry.Y = 9999
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnGetOdometry")
        End Try
    End Function


    Function ftnGetOdometry_Hero(ByRef tbpEstimatedPose As typeBotPose, ByRef tbpStartingPose As typeBotPose) As typeMapCoordinate ' returns a relative map coordinate
        Try
            Dim strRecv As String
            Dim strParsedText As String()
            ftnGetOdometry_Hero = tmcStartingOdometry
            strRecv = ftnGetHeroPosition(7)
            If CBool(InStr(strRecv, "ERROR")) Then strRecv = ftnGetHeroPosition(7)
            If CBool(InStr(strRecv, "ERROR")) Then strRecv = ftnGetHeroPosition(7)
            strParsedText = strRecv.Split(" "c)
            If (UBound(strParsedText) >= 0) And Not CBool(InStr(strRecv, "ERROR")) Then
                ftnGetOdometry_Hero.X = Round(Val(strParsedText(0)) / sngGridSize) * 16 / 14
                ftnGetOdometry_Hero.Y = tbpEstimatedPose.Location.Y - tbpStartingPose.Location.Y
                If boolDebugOdometry Then frmDebug.subWriteLogs(Color.Black, "     ---------   ftnGetOdometry_Hero  : " & ftnGetOdometry_Hero.X & ", " & ftnGetOdometry_Hero.Y)
            Else  ' Error - use dead reckoning instead
                If boolDebugOdometry Then frmDebug.subWriteLogs(Color.Red, Now() & " ftnGetOdometry_Hero error:  " & strRecv)
                ftnGetOdometry_Hero.X = 9999
                ftnGetOdometry_Hero.Y = 9999
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnGetOdometry_Hero")
        End Try
    End Function

    Function ftnGetOdometry_ER1(ByRef tbpEstimatedPose As typeBotPose, ByRef tbpStartingPose As typeBotPose) As typeMapCoordinate ' returns a relative map coordinate
        Try
            Dim strRecv As String
            Dim strParsedText As String()
            ftnGetOdometry_ER1 = tmcStartingOdometry
            strRecv = ftnSendCommand("position")
            If CBool(InStr(strRecv, "ERROR")) Then strRecv = ftnSendCommand("position")
            If CBool(InStr(strRecv, "ERROR")) Then strRecv = ftnSendCommand("position")
            strParsedText = strRecv.Split(" "c)
            If (UBound(strParsedText) = 2) And CBool(InStr(strRecv, "OK")) And Not CBool(InStr(strRecv, "ERROR")) Then
                ftnGetOdometry_ER1.X = Round(Val(strParsedText(1)) / sngGridSize)
                ftnGetOdometry_ER1.Y = Round(Val(strParsedText(2)) / sngGridSize)
                If boolDebugOdometry Then frmDebug.subWriteLogs(Color.Black, "     ---------   ftnGetOdometry_ER1  : " & ftnGetOdometry_ER1.X & ", " & ftnGetOdometry_ER1.Y)
            Else  ' Error - use dead reckoning instead
                If boolDebugOdometry Then frmDebug.subWriteLogs(Color.Red, Now() & " ftnGetOdometry_ER1 error:  " & strRecv)
                ftnGetOdometry_ER1.X = 9999
                ftnGetOdometry_ER1.Y = 9999
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnGetOdometry_ER1")
        End Try
    End Function


End Module
