﻿Imports System.Threading
Imports System.Math


Module modMapDraw

    '=================================================

    '       Map Draw MODULE
    '  This module is responsible for:
    '    - Drawing on the map (Bot, Destination, Path, etc)
    '

    '=================================================

    Public Sub subDrawBot(ByRef tbpBotPose As typeBotPose)    ' Draw the robot icon on the map.  The icon is a green circle with a yellow line to indicate orientation
        Try
            If Not GotABotClass.myMap.InvokeRequired Then
                If Not (tbpBotPose.Location.X = 0 And tbpBotPose.Location.Y = 0) Then
                    Dim sngDiameterZoomGrid As Single = sngBotDiameterGrid * sngZoom
                    tbpBotPose.Location = ftnCheckMapArrayLimits(tbpBotPose.Location)
                    Dim intBotRadius As Short = sngDiameterZoomGrid \ 2
                    Dim intZoomBotX As Short = Int(tbpBotPose.Location.X * sngZoom)
                    Dim intZoomBotY As Short = Int(tbpBotPose.Location.Y * sngZoom)
                    Dim sngOrientationRadians As Single = ftnDegreeToRadian(tbpBotPose.Orientation)
                    GotABotClass.myMap.CreateGraphics.DrawArc(Pens.Green, (intZoomBotX - intBotRadius), (intZoomBotY - intBotRadius), sngDiameterZoomGrid, sngDiameterZoomGrid, 0, 360)
                    GotABotClass.myMap.CreateGraphics.FillEllipse(Brushes.Green, (intZoomBotX - intBotRadius), (intZoomBotY - intBotRadius), sngDiameterZoomGrid, sngDiameterZoomGrid)
                    GotABotClass.myMap.CreateGraphics.DrawLine(Pens.Yellow, intZoomBotX, intZoomBotY, CInt(intZoomBotX + intBotRadius * Math.Cos(sngOrientationRadians)), CInt(intZoomBotY - intBotRadius * Math.Sin(sngOrientationRadians)))
                    If Not boolBotIsMoving Then subDrawBotUncertaintyOutline()
                    If (intHeadAngle <> 0) Then subDrawVectorFromBot(tbpBotPose, sngBotRadius, intHeadAngle, Color.White)
                    'GotABotClass.myMap.CreateGraphics.Flush()
                    'GotABotClass.myMap.CreateGraphics.Dispose()
                End If
                If frmLandMarkEdit.Visible Then
                    subDrawLandMark()
                    frmLandMarkEdit.Refresh()
                End If
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in subDrawBot")
        End Try
    End Sub


    Public Sub subEraseOldBotLocation(ByRef tmcOldBotLocation As typeMapCoordinate)
        Try
            If Not GotABotClass.InvokeRequired Then
                If boolShowMore Then
                    Dim sngDiameterZoomGrid As Single = sngBotDiameterGrid * sngZoom
                    Dim intBotRadius As Integer = sngDiameterZoomGrid \ 2
                    Dim intZoomBotX As Integer = Int(tmcOldBotLocation.X * sngZoom) - intBotRadius
                    Dim intZoomBotY As Integer = Int(tmcOldBotLocation.Y * sngZoom) - intBotRadius
                    tmcOldBotLocation = ftnCheckMapArrayLimits(tmcOldBotLocation)
                    If Not (tmcOldBotLocation.X = 0 And tmcOldBotLocation.Y = 0) Then
                        GotABotClass.myMap.CreateGraphics.DrawArc(Pens.White, intZoomBotX, intZoomBotY, sngDiameterZoomGrid, sngDiameterZoomGrid, 0, 360)
                        GotABotClass.myMap.CreateGraphics.FillEllipse(Brushes.White, intZoomBotX, intZoomBotY, sngDiameterZoomGrid, sngDiameterZoomGrid)
                    End If
                End If
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in subEraseOldBotLocation")
        End Try
    End Sub

    Public Sub subDrawBotUncertaintyOutline()
        Try
            If boolShowMore And Not GotABotClass.InvokeRequired Then
                Dim sngHalfDiameterZoomGrid As Single = sngBotRadiusGrid * sngZoom
                Dim intZoomBotX As Short = Int(tbpBotPose.Location.X * sngZoom)
                Dim intZoomBotY As Short = Int(tbpBotPose.Location.Y * sngZoom)
                Dim LocX, LocY As Short
                Dim LocY2, LocX2 As Short
                Dim sngLeftUncertaintyRadians As Single = ftnDegreeToRadian(tbpBotPose.Orientation + tbpBotPose.OrientationUncertainty)
                Dim sngRightUncertaintyRadians As Single = ftnDegreeToRadian(tbpBotPose.Orientation - tbpBotPose.OrientationUncertainty)
                LocX = (tbpBotPose.Location.X - tbpBotPose.LocationUncertainty.X)
                LocY = (tbpBotPose.Location.Y - tbpBotPose.LocationUncertainty.Y)
                subCheckMapArrayLimits(LocX, LocY)
                LocX2 = (tbpBotPose.Location.X + tbpBotPose.LocationUncertainty.X)
                LocY2 = (tbpBotPose.Location.Y + tbpBotPose.LocationUncertainty.Y)
                subCheckMapArrayLimits(LocX2, LocY2)

                GotABotClass.myMap.CreateGraphics.DrawRectangle(Pens.YellowGreen, LocX * sngZoom, (LocY * sngZoom), _
                                                                       ((LocX2 - LocX) * sngZoom), ((LocY2 - LocY) * sngZoom))
                GotABotClass.myMap.CreateGraphics.DrawLine(Pens.YellowGreen, intZoomBotX, intZoomBotY, _
                                                           CInt(intZoomBotX + sngHalfDiameterZoomGrid * Math.Cos(sngLeftUncertaintyRadians)), _
                                                           CInt(intZoomBotY - sngHalfDiameterZoomGrid * (Math.Sin(sngLeftUncertaintyRadians))))
                GotABotClass.myMap.CreateGraphics.DrawLine(Pens.YellowGreen, intZoomBotX, (tbpBotPose.Location.Y * sngZoom), _
                                                           CInt(intZoomBotX + sngHalfDiameterZoomGrid * (Math.Cos(sngRightUncertaintyRadians))), _
                                                           CInt(intZoomBotY - sngHalfDiameterZoomGrid * (Math.Sin(sngRightUncertaintyRadians))))
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in subDrawBotUncertaintyOutline")
        End Try
    End Sub



    Public Sub subDrawDestination(ByRef tmcDestination As typeMapCoordinate)    ' Draw the destination icon on the map.  The icon is a red 'X' (as in "X marks the spot and red indicates STOP here)
        Try
            If Not GotABotClass.InvokeRequired Then
                Dim wideRedPen As New Pen(Brushes.Red)
                Dim intBotRadius As Integer = (sngBotRadiusGrid * sngZoom)
                tmcDestination = ftnCheckMapArrayLimits(tmcDestination)
                Dim tmcZoomDestinationX As Integer = tmcDestination.X * sngZoom
                Dim tmcZoomDestinationY As Integer = tmcDestination.Y * sngZoom
                wideRedPen.Width = 8.0F
                wideRedPen.LineJoin = Drawing2D.LineJoin.Bevel
                GotABotClass.myMap.CreateGraphics.DrawLine(wideRedPen, tmcZoomDestinationX - intBotRadius, tmcZoomDestinationY - intBotRadius, tmcZoomDestinationX + intBotRadius, tmcZoomDestinationY + intBotRadius)
                GotABotClass.myMap.CreateGraphics.DrawLine(wideRedPen, tmcZoomDestinationX - intBotRadius, tmcZoomDestinationY + intBotRadius, tmcZoomDestinationX + intBotRadius, tmcZoomDestinationY - intBotRadius)
                wideRedPen.Dispose()
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in subDrawDestination")
        End Try
    End Sub

    Public Sub subDrawVectorFromBot(ByRef tbpEstimatedPose As typeBotPose, ByRef sngRange As Single, ByRef sngRelativeAngle As Single, ByRef boolGreen As Boolean)    ' Draw a line from the bot to the Ranger supplied vector
        Try
            If boolShowMore And Not GotABotClass.InvokeRequired Then
                Dim tmcEndPoint As typeMapCoordinate = ftnLocationFromEstimatedPose(tbpEstimatedPose, sngRelativeAngle, sngRange)
                tmcEndPoint = ftnCheckMapArrayLimits(tmcEndPoint)
                If boolGreen Then
                    GotABotClass.myMap.CreateGraphics.DrawLine(Pens.Lime, (tbpBotPose.Location.X * sngZoom), (tbpBotPose.Location.Y * sngZoom), (tmcEndPoint.X * sngZoom), (tmcEndPoint.Y * sngZoom))
                    subDrawSmallCircle(tmcEndPoint, Color.Lime)
                Else
                    GotABotClass.myMap.CreateGraphics.DrawLine(Pens.Red, (tbpBotPose.Location.X * sngZoom), (tbpBotPose.Location.Y * sngZoom), (tmcEndPoint.X * sngZoom), (tmcEndPoint.Y * sngZoom))
                    subDrawSmallCircle(tmcEndPoint, Color.Red)
                End If
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in subDrawVectorFromBot")
        End Try
    End Sub

    Public Sub subDrawVectorFromBot(ByRef tbpEstimatedPose As typeBotPose, ByRef sngRange As Single, ByRef sngRelativeAngle As Single, ByRef colVector As Color)    ' Draw a line from the bot to the Ranger supplied vector
        Try
            If boolShowMore And Not GotABotClass.InvokeRequired Then
                Dim tmcEndPoint As typeMapCoordinate = ftnLocationFromEstimatedPose(tbpEstimatedPose, sngRelativeAngle, sngRange)
                tmcEndPoint = ftnCheckMapArrayLimits(tmcEndPoint)
                Dim myPen As Pen
                myPen = New Pen(colVector)
                GotABotClass.myMap.CreateGraphics.DrawLine(myPen, (tbpBotPose.Location.X * sngZoom), (tbpBotPose.Location.Y * sngZoom), (tmcEndPoint.X * sngZoom), (tmcEndPoint.Y * sngZoom))
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in subDrawVectorFromBot")
        End Try
    End Sub

    Public Sub subDrawVector(ByRef tmcStartLocation As typeMapCoordinate, ByRef trvVector As typeRangeVector, ByRef boolGreen As Boolean)    ' Draw a vector
        Try
            If boolShowMore And Not GotABotClass.InvokeRequired Then
                Dim tmcEndPoint As typeMapCoordinate = ftnLocationFromLocation(tmcStartLocation, trvVector.sngAngle, trvVector.sngRange)
                If boolGreen Then
                    GotABotClass.myMap.CreateGraphics.DrawLine(Pens.Lime, (tmcStartLocation.X * sngZoom), (tmcStartLocation.Y * sngZoom), (tmcEndPoint.X * sngZoom), (tmcEndPoint.Y * sngZoom))
                    subDrawSmallCircle(tmcEndPoint, Color.Lime)
                Else
                    GotABotClass.myMap.CreateGraphics.DrawLine(Pens.Red, (tmcStartLocation.X * sngZoom), (tmcStartLocation.Y * sngZoom), (tmcEndPoint.X * sngZoom), (tmcEndPoint.Y * sngZoom))
                    subDrawSmallCircle(tmcEndPoint, Color.Red)
                End If
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in subDrawVector")
        End Try
    End Sub

    Public Sub subDrawVector(ByRef tmcStartLocation As typeMapCoordinate, ByRef sngRange As Single, ByRef sngOrientation As Single, ByRef boolGreen As Boolean)    ' Draw a line from the Range Finder
        Try
            If boolShowMore And Not GotABotClass.InvokeRequired Then
                Dim tmcEndPoint As typeMapCoordinate = ftnLocationFromLocation(tmcStartLocation, sngOrientation, sngRange)
                If boolGreen Then
                    GotABotClass.myMap.CreateGraphics.DrawLine(Pens.Lime, (tmcStartLocation.X * sngZoom), (tmcStartLocation.Y * sngZoom), (tmcEndPoint.X * sngZoom), (tmcEndPoint.Y * sngZoom))
                    subDrawSmallCircle(tmcEndPoint, Color.Lime)
                Else
                    GotABotClass.myMap.CreateGraphics.DrawLine(Pens.Red, (tmcStartLocation.X * sngZoom), (tmcStartLocation.Y * sngZoom), (tmcEndPoint.X * sngZoom), (tmcEndPoint.Y * sngZoom))
                    subDrawSmallCircle(tmcEndPoint, Color.Red)
                End If
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in subDrawVector")
        End Try
    End Sub

    Public Sub subDrawBotToLM(ByVal tmcLM As typeMapCoordinate, ByVal colColor As Color)    ' Draw a line from the bot to the LM
        Try
            If boolShowMore And Not GotABotClass.InvokeRequired Then
                Dim myPen As Pen
                myPen = New Pen(colColor)
                tmcLM = ftnCheckMapArrayLimits(tmcLM)
                GotABotClass.myMap.CreateGraphics.DrawLine(myPen, (tbpBotPose.Location.X * sngZoom), (tbpBotPose.Location.Y * sngZoom), (tmcLM.X * sngZoom), (tmcLM.Y * sngZoom))
                If frmLandMarkEdit.Visible Then
                    subDrawLandMark()
                    frmLandMarkEdit.Refresh()
                End If
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in subDrawBotToLM")
        End Try
    End Sub


    Public Sub subDrawLandMark()
        Try
            If Not GotABotClass.InvokeRequired Then
                Dim LocX, LocY As Single
                Dim ValidX, ValidY, ValidHeight, ValidWidth As Single
                LocX = Convert.ToSingle(frmLandMarkEdit.UpDnLocX.Value)
                LocY = Convert.ToSingle(frmLandMarkEdit.UpDnLocY.Value)
                subCheckMapArrayLimits(LocX, LocY)
                ValidX = Convert.ToSingle(frmLandMarkEdit.UpDnValidX.Value)
                ValidY = Convert.ToSingle(frmLandMarkEdit.UpDnValidY.Value)
                subCheckMapArrayLimits(ValidX, ValidY)
                ValidWidth = Convert.ToSingle(frmLandMarkEdit.UpDnValidWidth.Value)
                ValidHeight = Convert.ToSingle(frmLandMarkEdit.UpDnValidHeight.Value)

                Dim sngDotHeight As Single = 4 * sngZoom
                GotABotClass.myMap.CreateGraphics.DrawArc(Pens.CornflowerBlue, (LocX * sngZoom - sngDotHeight / 2), _
                                                     (LocY * sngZoom - sngDotHeight / 2), sngDotHeight, sngDotHeight, 0, 360)
                GotABotClass.myMap.CreateGraphics.FillEllipse(Brushes.CornflowerBlue, (LocX * sngZoom - sngDotHeight / 2), _
                                                                LocY * sngZoom - sngDotHeight / 2, sngDotHeight, sngDotHeight)
                GotABotClass.myMap.CreateGraphics.DrawRectangle(Pens.CornflowerBlue, ValidX * sngZoom, (ValidY * sngZoom), _
                                                                       (ValidWidth * sngZoom), (ValidHeight * sngZoom))
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in subDrawLandMark")
        End Try
    End Sub

    Public Sub subDrawMapPixel(ByRef intPixelX As Short, ByRef intPixelY As Short, ByRef colPixel As Color)
        Try
            If boolShowMore And Not GotABotClass.InvokeRequired Then
                subCheckMapArrayLimits(intPixelX, intPixelY)
                Dim myBrush As Brush
                myBrush = New SolidBrush(colPixel)
                Dim myPen As Pen
                myPen = New Pen(colPixel)
                Dim intZoomPixelX As Short = intPixelX * sngZoom - sngZoom
                Dim intZoomPixelY As Short = intPixelY * sngZoom - sngZoom
                GotABotClass.myMap.CreateGraphics.DrawRectangle(myPen, intZoomPixelX, intZoomPixelY, sngZoom, sngZoom)
                GotABotClass.myMap.CreateGraphics.FillRectangle(myBrush, intZoomPixelX, intZoomPixelY, sngZoom, sngZoom)
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in subDrawMapPixel")
        End Try
    End Sub

    Public Sub subShowPath(ByVal tmcDestination As typeMapCoordinate)  ' Draw the Bot to Destination path on the map
        Try
            If boolShowMore And Not GotABotClass.InvokeRequired Then
                Dim tmcPathStep As typeMapCoordinate
                Dim tmcOldPathStep As typeMapCoordinate
                Dim boolNearWall As Boolean = False

                tmcPathStep.X = CShort(Int(tbpBotPose.Location.X))
                tmcPathStep.Y = CShort(Int(tbpBotPose.Location.Y))
                tmcOldPathStep = tbpBotPose.Location
                tmcPathStep = ftnLowestValueNeighbor(tmcPathStep)
                subHighLightPathStep(tmcPathStep.X, tmcPathStep.Y)
                Do While ((intPathArray(tmcPathStep.X, tmcPathStep.Y) <> constDestinationCost) And _
                          (tmcPathStep.X <> tmcOldPathStep.X) Or (tmcPathStep.Y <> tmcOldPathStep.Y)) _
                          And Not boolStop And Not boolShutDown '
                    tmcOldPathStep = tmcPathStep
                    tmcPathStep = ftnLowestValueNeighbor(tmcPathStep)
                    subHighLightPathStep(tmcPathStep.X, tmcPathStep.Y)
                Loop
                subDrawBot(tbpBotPose) ' Restore the bot icon
                subDrawDestination(tmcDestination) ' Restore the destination icon
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in subShowPath")
        End Try
    End Sub



    Public Sub subHighLightPathStep(ByRef intPathX As Short, ByRef intPathY As Short)
        Try
            Dim sngZoomGrid As Single = sngZoom / sngGridSize
            Dim intBotRadius As Integer = (sngBotDiameter * sngZoomGrid) \ 2
            Dim intX_Left As Short = intPathX * sngZoom - intBotRadius  ' TODO intBotRadiusGrid???
            Dim intY_Left As Short = intPathY * sngZoom - intBotRadius
            Dim intBotDiameterZoom As Integer = sngBotDiameter * sngZoomGrid
            GotABotClass.myMap.CreateGraphics.DrawArc(Pens.Yellow, _
                                intX_Left, intY_Left, _
                                intBotDiameterZoom, intBotDiameterZoom, 0, 360)
            GotABotClass.myMap.CreateGraphics.FillEllipse(Brushes.Yellow, _
                             intX_Left, intY_Left, _
                            intBotDiameterZoom, intBotDiameterZoom)
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in subHighLightPathStep")
        End Try
    End Sub





    Delegate Sub DelegateDrawSmallCircle(ByVal tmcMapLocation As typeMapCoordinate, ByVal colCircle As Color)
    Public Sub subDrawSmallCircle(ByVal tmcMapLocation As typeMapCoordinate, ByVal colCircle As Color)    ' Draw a small  circle 
        Try
            Dim myBrush As Brush
            myBrush = New SolidBrush(colCircle)
            Dim myPen As Pen
            myPen = New Pen(colCircle)
            If GotABotClass.InvokeRequired And Not boolShutDown And boolShowMore Then
                Dim newDelegate As New DelegateDrawSmallCircle(AddressOf subDrawSmallCircle)
                GotABotClass.Invoke(newDelegate, New Object() {tmcMapLocation})
                Thread.Sleep(50)
            ElseIf boolShowMore And Not GotABotClass.InvokeRequired Then
                If Not (tmcMapLocation.X = 0 And tmcMapLocation.Y = 0) Then
                    tmcMapLocation = ftnCheckMapArrayLimits(tmcMapLocation)
                    Dim intX_Location As Short = (tmcMapLocation.X - 2.5) * sngZoom
                    Dim intY_Location As Short = (tmcMapLocation.Y - 2.5) * sngZoom
                    Dim intWidthOrHeight As Short = 5 * sngZoom
                    GotABotClass.myMap.CreateGraphics.DrawArc(myPen, intX_Location, intY_Location, intWidthOrHeight, intWidthOrHeight, 0, 360)
                    GotABotClass.myMap.CreateGraphics.FillEllipse(myBrush, intX_Location, intY_Location, intWidthOrHeight, intWidthOrHeight)
                End If
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in subDrawSmallCircle")
        End Try
    End Sub



    Public Sub subDrawBumperSwitch(ByVal tbpEstimatedPose As typeBotPose, ByVal sngRelativeAngle As Single)    ' Highlight the bumper switch 
        Try
            If boolShowMore And Not GotABotClass.InvokeRequired Then
                Dim tmcBumperSwitch As typeMapCoordinate
                Dim sngDiameter As Single = sngBotRadiusGrid / 2 * sngZoom
                tmcBumperSwitch = ftnLocationFromEstimatedPose(tbpEstimatedPose, sngRelativeAngle, sngBotRadius)
                If Not (tmcBumperSwitch.X = 0 And tmcBumperSwitch.Y = 0) Then
                    GotABotClass.myMap.CreateGraphics.DrawArc(Pens.Red, (tmcBumperSwitch.X * sngZoom - sngBotRadius), (tmcBumperSwitch.Y * sngZoom - sngBotRadius), sngDiameter, sngDiameter, 0, 360)
                    GotABotClass.myMap.CreateGraphics.FillEllipse(Brushes.Red, (tmcBumperSwitch.X * sngZoom - sngBotRadius), (tmcBumperSwitch.Y * sngZoom - sngBotRadius), sngDiameter, sngDiameter)
                End If
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in subDrawBumperSwitch")
        End Try
    End Sub


    Public Sub subDrawScanVectors(ByRef tbpPose As typeBotPose, ByRef boolGreen As Boolean)
        Try
            If boolShowMore And Not boolBotIsMoving And Not GotABotClass.InvokeRequired Then
                If ftnScannerEnabled() Then
                    Dim tmcVectorOrigin As typeMapCoordinate = ftnLocationFromEstimatedPose(tbpPose, tmcScanVectorOffset)
                    Dim intStep As Short = Max(constMaxScannerBeams \ intNumberOfBeamsDefault, 2)
                    For intBeamNumber As Short = 0 To constMaxScannerBeams - 1 Step intStep ' Draw Scan Vectors in Green or Red
                        If (trvScanVectors(intBeamNumber).sngRange > 0) Then _
                                subDrawVector(tmcVectorOrigin, trvScanVectors(intBeamNumber).sngRange, _
                                               ftnConvertTo360(tbpPose.Orientation + trvScanVectors(intBeamNumber).sngAngle), boolGreen) ' Show Scan Vector on map
                    Next
                End If
                If boolRangersAreEnabled And boolGreen Then ' Draw Ranger Vectors  
                    Dim trvRangerBeams(intHighestEnabledRangerNumber) As typeRangeVector  ' Range vector - relative to Bot center
                    For intRangerNumber As Short = 1 To intHighestEnabledRangerNumber ' 
                        If trfRangeFinders(intRangerNumber).boolEnable Then
                            trvRangerBeams(intRangerNumber) = frmRangeFinder.ftnRangeFinderVector(tbpPose, intRangerNumber, trfRangeFinders(intRangerNumber).sngRange, trfRangeFinders(intRangerNumber).intOrientation) ' Find Range vector - relative to Bot center
                            subDrawVector(tbpPose.Location, trvRangerBeams(intRangerNumber).sngRange, _
                                ftnConvertTo360(tbpPose.Orientation + trvRangerBeams(intRangerNumber).sngAngle), boolGreen)
                        End If
                    Next
                End If
                If boolGreen Then subWait(100)
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in subDrawScanVectors")
        End Try
    End Sub



End Module
