﻿Imports System.Math

Module modPath
    Const constNearWall = 512

    '=================================================

    '  This module is responsible for:
    '    - operates on map array generated by the modNavigationArray module
    '    - Generates a path (shortest distance) between the bot and the destination
    '

    '=================================================

    '  Convert the map array to a Path Matrix where the Destination is assigned the lowest value (ie Cost) 
    '  and the robot's location is assigned a much higher value.  The shortest path between robot and Destination may 
    ' be found by following the gradient.  i.e. Starting at the Robot's location, find the neighboring cell with the lowest value.
    ' This will be the next cell on the path to the destination.  Repeat until the destination is reached.


    Sub subGeneratePathArray()
        '  This routine generates a matrix representing the obstacles in the map
        Try
            GotABotClass.Cursor = Cursors.WaitCursor
            ReDim intPathArray(CInt(imgMap.Width), CInt(imgMap.Height))
            strStatusDriving = intDrivingSequenceNumber & " : Generating Path Matrix "
            Dim intX, intY As Short
            Dim intDistanceFromWallSquaredCell As Short
            For intX = 0 To imgMap.Width - 0
                For intY = 0 To imgMap.Height - 0
                    intDistanceFromWallSquaredCell = intDistanceFromWallSquaredArray(intX, intY)
                    If boolOccupiedCell(intX, intY) Then
                        intPathArray(intX, intY) = 65535
                    ElseIf (intDistanceFromWallSquaredCell < sngBotRadiusSquared) Then ' 
                        intPathArray(intX, intY) = constMaxPathCost - 4096   '    ' 
                    ElseIf (intDistanceFromWallSquaredCell < intMaxDistanceFromWallSquared) Then ' 
                        intPathArray(intX, intY) = Sqrt(intMaxDistanceFromWallSquared - intDistanceFromWallSquaredCell)   '    ' 
                    ElseIf boolGreySpace(intX, intY) Then  ' 
                        intPathArray(intX, intY) = 8
                    Else
                        intPathArray(intX, intY) = 0
                    End If
                Next intY
            Next intX
            GotABotClass.Cursor = Cursors.Arrow
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in subGeneratePathArray")
        End Try
    End Sub

    Dim boolCostIsUsed(constMaxPathCost) As Boolean  ' speed up ftnGeneratePathCosts by identifying unused intNextLocationCost values
    Dim intMaxThreshold As UShort
    Function ftnGeneratePathCosts(ByRef tmcDestination As typeMapCoordinate, ByRef tmcStartLocation As typeMapCoordinate) As Boolean
        ' Returns true if a good path is found
        Try
            GotABotClass.Cursor = Cursors.WaitCursor
            ftnGeneratePathCosts = True
            strStatusDriving = intDrivingSequenceNumber & " : Finding Path to Destination "
            Dim intX, intY As Short
            Dim intX_Max As Short = UBound(intPathArray, 1) - 1
            Dim intY_Max As Short = UBound(intPathArray, 2) - 1
            Dim intCurrentLocationCost As UShort = constDestinationCost + 1 ' the destination is the lowest cost. constDestinationCost = 4096
            Dim intNextLocationCost As UShort
            intMaxThreshold = constNearWall
            If (ftnInWall(tbpBotPose.Location)) Then intMaxThreshold += 1

            ReDim boolCostIsUsed(constMaxPathCost) ' set all elements to False
            intPathArray(tmcDestination.X, tmcDestination.Y) = intCurrentLocationCost ' Start path at destination = lowest cost

            Dim timeOut As DateTime = Now().AddMilliseconds(30000)
            While (intPathArray(tmcStartLocation.X, tmcStartLocation.Y) < constDestinationCost) And Not boolStop And (intCurrentLocationCost <> constMaxPathCost)
                intNextLocationCost = intCurrentLocationCost + 1
                For intX = 0 To intX_Max ' Work backward from the destination to the bot.  Each iteration (Location), add to the cost
                    For intY = 0 To intY_Max
                        If ((intPathArray(intX, intY) = intCurrentLocationCost)) And Not boolStop Then
                            subUpdateCellNeighbors(intX, intY, intNextLocationCost)
                        End If
                    Next intY
                    System.Windows.Forms.Application.DoEvents()
                Next intX
                intCurrentLocationCost += 1
                While Not boolCostIsUsed(intCurrentLocationCost) And (intCurrentLocationCost <> constMaxPathCost)
                    intCurrentLocationCost += 1
                End While
                If boolSlam Then
                    If (Now() > timeOut) Then
                        frmDebug.subWriteLogs(Color.Green, " TimeOut in ftnGeneratePathCosts")
                        Exit While ' structured this way for speed, in non Slam mode only boolSlam needs to be evaluated not Now()
                    End If
                End If
            End While
            If (intCurrentLocationCost = constMaxPathCost) Or (boolSlam And (Now() > timeOut)) Then
                ftnGeneratePathCosts = False
                If Not boolSlam Then
                    strStatusDriving = intDrivingSequenceNumber & " : Invalid Destination. -"
                    GotABotClass.subUpdateAllText(tbpBotPose)
                    ftnSendCommand("play phrase 'Error - In valid Destination'")
                End If
            End If
            GotABotClass.Cursor = Cursors.Arrow
        Catch ex As Exception
            GotABotClass.Cursor = Cursors.Arrow
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnGeneratePathCosts")
        End Try
    End Function

    Sub subUpdateCellNeighbors(ByRef intX As Short, ByRef intY As Short, ByVal intNextLocationCost As UShort)
        '  Up/Dn and Left/Right neighbors of the selected cell are incremented by +1.
        '  Diagonal Neighbors are incremented by +2
        Try
            Dim intCellX, intCellY As Short
            Dim intCellX_Max As Short = UBound(intPathArray, 1)
            Dim intCellY_Max As Short = UBound(intPathArray, 2)

            intCellX = intX
            For intCellY = (intY - 1) To (intY + 1) Step 2 ' up and down neighbors
                If ((intCellX <> -1) And (intCellY <> -1) And intCellY <> intCellY_Max) Then _
                    subUpdateCell(intCellX, intCellY, intNextLocationCost)
            Next intCellY
            intCellY = intY
            For intCellX = (intX - 1) To (intX + 1) Step 2 ' left and right neighbors
                If ((intCellX <> -1) And intCellX <> intCellX_Max And (intCellY <> -1)) Then _
                    subUpdateCell(intCellX, intCellY, intNextLocationCost)
            Next intCellX

            intNextLocationCost += 1  ' cost = 2 for diagonal neighbors
            For intCellX = (intX - 1) To (intX + 1) Step 2 ' diagonals
                For intCellY = (intY - 1) To (intY + 1) Step 2 ' 
                    If ((intCellX <> -1) And intCellX <> intCellX_Max And (intCellY <> -1) And intCellY <> intCellY_Max) Then _
                      subUpdateCell(intCellX, intCellY, intNextLocationCost)
                Next intCellY
            Next intCellX
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in subUpdateCellNeighbors")
        End Try
    End Sub


    Private Sub subUpdateCell(ByRef intCellX As Short, ByRef intCellY As Short, ByRef intNextLocationCost As UShort)
        Try
            If (intPathArray(intCellX, intCellY) <= intMaxThreshold) Then  ' if intPathArray has not been previously set
                intPathArray(intCellX, intCellY) += intNextLocationCost
                boolCostIsUsed(intPathArray(intCellX, intCellY)) = True
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in subUpdateCell")
        End Try
    End Sub

    Public Function ftnLowestValueNeighbor(ByRef tmcCurrentLocation As typeMapCoordinate) As typeMapCoordinate
        ' This function finds the next Location on the path as we move
        ' from bot to destination.  The next Location is the neighbor of tmcCurrentLocation
        ' which has the lowest intPathArray value).
        Try
            ftnLowestValueNeighbor = tmcCurrentLocation
            Dim tmcCandidateLocation As typeMapCoordinate
            Dim intLowestValue As UShort = intPathArray(ftnLowestValueNeighbor.X, ftnLowestValueNeighbor.Y)
            Dim intCandidateValue As UShort
            For tmcCandidateLocation.X = (tmcCurrentLocation.X - 1) To (tmcCurrentLocation.X + 1)
                For tmcCandidateLocation.Y = (tmcCurrentLocation.Y - 1) To (tmcCurrentLocation.Y + 1)
                    intCandidateValue = intPathArray(tmcCandidateLocation.X, tmcCandidateLocation.Y)
                    If ((intCandidateValue >= constDestinationCost) And (intCandidateValue < intLowestValue)) Then
                        ftnLowestValueNeighbor = tmcCandidateLocation
                        intLowestValue = intCandidateValue
                    End If
                Next tmcCandidateLocation.Y
            Next tmcCandidateLocation.X
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnLowestValueNeighbor")
        End Try
    End Function



    '==============================================================================================
    '   The following functions are used to find the intermediate steps along the path from a start location
    '    to the Destination
    '==============================================================================================

    Function ftnFindTargetPose(ByVal tbpEstimatedPose As typeBotPose, ByRef tmcDestination As typeMapCoordinate) As typeBotPose
        ' This function searches for the furthest point on the path that it can safely reach using
        ' an optional 'turn' command followed by a move 'fwd' or 'MoveXY'command.  
        ftnFindTargetPose = tbpEstimatedPose
        Try
            GotABotClass.Cursor = Cursors.WaitCursor
            If boolDebugPath Then
                frmDebug.subWriteLogs(Color.Black, ftnElapsedTime(timeMoveCommandSent) & " ftnFindTargetPose Start ")
                frmDebug.subWriteLogs(Color.Black, "             tbpEstimatedPose= " & tbpEstimatedPose.Location.X & ", " & tbpEstimatedPose.Location.Y & "  at " & tbpEstimatedPose.Orientation)
                frmDebug.subWriteLogs(Color.Black, "             tmcDestination= " & tmcDestination.X & ", " & tmcDestination.Y)
            End If
            Dim boolClearPath As Boolean = False
            Dim strCurrentTypeOfMove As String = "Fwd"
            Dim tbpFurthestSafePose As typeBotPose = tbpEstimatedPose
            Do           'see if we can go furthur without hitting a wall
                boolClearPath = False
                tbpFurthestSafePose = ftnFindTargetPose
                ftnFindTargetPose.Location = ftnLowestValueNeighbor(ftnFindTargetPose.Location)
                ftnFindTargetPose.Orientation = ftnP1ToP2Angle(tbpEstimatedPose.Location, ftnFindTargetPose.Location) ' Orientation required to drive from tbpEstimatedPose.Location to ftnFindTargetPose
                tbpEstimatedPose.Orientation = ftnFindTargetPose.Orientation
                If (strCurrentTypeOfMove = "Fwd") Then _
                    boolClearPath = ftnClearPathFwd(tbpEstimatedPose, ftnFindTargetPose.Location) ' can we drive forward to this location without coming too close to a wall?
                If Not boolClearPath And boolEnableXY_Motion Then
                    strCurrentTypeOfMove = "XY"
                    boolClearPath = ftnMoveXY_ConditionsMet(tbpEstimatedPose, ftnFindTargetPose.Location)
                End If
                subDoEventsIfMainThread()
                If boolDebugPath Then
                    frmDebug.subWriteLogs(Color.Black, ": ftnFindTargetPose: tbpFurthestSafePose= " & tbpFurthestSafePose.Location.X & ", " & tbpFurthestSafePose.Location.Y & "  at " & tbpFurthestSafePose.Orientation)
                    frmDebug.subWriteLogs(Color.Black, ": ftnFindTargetPose: ftnFindTargetPose= " & ftnFindTargetPose.Location.X & ", " & ftnFindTargetPose.Location.Y & "  at " & ftnFindTargetPose.Orientation)
                    frmDebug.subWriteLogs(Color.Black, ": ftnFindTargetPose: tbpEstimatedPose= " & tbpEstimatedPose.Location.X & ", " & tbpEstimatedPose.Location.Y & "  at " & tbpEstimatedPose.Orientation)
                    frmDebug.subWriteLogs(Color.Black, ": ftnFindTargetPose: strCurrentTypeOfMove " & strCurrentTypeOfMove & ":  boolClearPath = " & boolClearPath)
                    frmDebug.subWriteLogs(Color.Black, ": ")
                End If
            Loop Until (boolStop Or Not boolClearPath) Or ftnLocationsEqual(ftnFindTargetPose.Location, tmcDestination) Or ftnLocationsEqual(ftnFindTargetPose.Location, tbpFurthestSafePose.Location)
            ftnFindTargetPose = tbpFurthestSafePose
            GotABotClass.Cursor = Cursors.Arrow
        Catch ex As Exception
            GotABotClass.Cursor = Cursors.Arrow
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnFindTargetPose")
        End Try
    End Function


    Function ftnTypeOfMove(ByRef tbpEstimatedPose As typeBotPose, ByRef tbpTargetPose As typeBotPose) As String
        Try
            If ftnClearPathFwd(tbpEstimatedPose, tbpTargetPose.Location) Then
                ftnTypeOfMove = "Fwd"
            ElseIf ftnMoveXY_ConditionsMet(tbpEstimatedPose, tbpTargetPose.Location) Then
                ftnTypeOfMove = "XY"
            ElseIf ftnTargetRequiresTurn(tbpEstimatedPose, tbpTargetPose.Location) Then
                ftnTypeOfMove = "Turn"
            Else
                frmDebug.subWriteLogs(Color.Green, " Unsafe move detected in ftnTypeOfMove")
                ftnTypeOfMove = "Turn"
            End If
        Catch ex As Exception
            ftnTypeOfMove = "Turn"
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnTypeOfMove")
        End Try
    End Function


    Function ftnTargetRequiresTurn(ByRef tbpEstimatedPose As typeBotPose, ByVal tmcTargetLocation As typeMapCoordinate) As Boolean
        Try
            ftnTargetRequiresTurn = False
            Dim tbpTargetPose As typeBotPose = tbpEstimatedPose
            tbpTargetPose.Location = tmcTargetLocation
            Dim sngAbsTurnAngle As Single = Abs(ftnTurnAngle(tbpEstimatedPose, tbpTargetPose))
            If (sngAbsTurnAngle > 60) Then
                ftnTargetRequiresTurn = True
            Else
                If ftnLocationsEqual(tbpEstimatedPose.Location, tmcTargetLocation) Then
                    ftnTargetRequiresTurn = True
                Else
                    Dim sngDistanceToTarget As Single = ftnDistanceBetweenLocations(tbpEstimatedPose.Location, tmcTargetLocation)
                    Dim sngOffsetDistance As Single = sngDistanceToTarget * Sin(ftnDegreeToRadian(sngAbsTurnAngle))
                    If Not boolSlam And (sngOffsetDistance >= sngOneInch) Then
                        If (sngOffsetDistance > sngTwoInches) Then
                            ftnTargetRequiresTurn = True
                        Else
                            tmcTargetLocation = ftnLocationFromEstimatedPose(tbpEstimatedPose, 0, sngDistanceToTarget)
                            If Not ftnClearPathFwd_NoTurnCheck(tbpEstimatedPose, tmcTargetLocation) Then _
                            ftnTargetRequiresTurn = True
                        End If
                    ElseIf boolSlam And (sngOffsetDistance >= sngTwoInches) Then
                        If (sngOffsetDistance > sngThreeInches) Then
                            ftnTargetRequiresTurn = True
                        Else
                            tmcTargetLocation = ftnLocationFromEstimatedPose(tbpEstimatedPose, 0, sngDistanceToTarget)
                            If Not ftnClearPathFwd_NoTurnCheck(tbpEstimatedPose, tmcTargetLocation) Then _
                            ftnTargetRequiresTurn = True
                        End If
                    End If
                End If
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnTargetRequiresTurn")
        End Try
    End Function


    Function ftnClearPathFwd(ByRef tbpEstimatedPose As typeBotPose, ByRef tmcTargetLocation As typeMapCoordinate) As Boolean
        ' this function returns true if the path between tbpEstimatedPose.Location and tmcTargetLocation is clear of obstructions
        ' and not near a wall
        Try
            ftnClearPathFwd = True
            If ftnTargetRequiresTurn(tbpEstimatedPose, tmcTargetLocation) Then
                ftnClearPathFwd = False
            Else
                Dim sngDistanceToTarget As Single = ftnDistanceBetweenLocations(tbpEstimatedPose.Location, tmcTargetLocation)
                If (sngDistanceToTarget > (sngBotRadius + sngTwoInches)) Then
                    Dim tmcCheckLocation As typeMapCoordinate = ftnLocationFromEstimatedPose(tbpEstimatedPose, 0, sngBotRadius) ' tbpEstimatedPose.Location 
                    tmcCheckLocation = ftnFirstPointOnLine(tmcCheckLocation, tmcTargetLocation)
                    Do
                        tmcCheckLocation = ftnNextPointOnLine(tmcCheckLocation, tmcTargetLocation)
                        ftnClearPathFwd = ftnSafeLocationForMove(tbpEstimatedPose, tmcCheckLocation, sngDistanceToTarget)
                    Loop Until ftnLocationsEqual(tmcCheckLocation, tmcTargetLocation) Or boolStop Or Not ftnClearPathFwd
                End If
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnClearPathFwd")
        End Try
    End Function

    Function ftnClearPathFwd_NoTurnCheck(ByRef tbpEstimatedPose As typeBotPose, ByRef tmcTargetLocation As typeMapCoordinate) As Boolean
        ' this function returns true if the path between tbpEstimatedPose.Location and tmcTargetLocation is clear of obstructions
        ' and not near a wall
        Try
            ftnClearPathFwd_NoTurnCheck = True
            Dim sngDistanceToTarget As Single = ftnDistanceBetweenLocations(tbpEstimatedPose.Location, tmcTargetLocation)
            If (sngDistanceToTarget > sngBotRadius) Then
                Dim tmcCheckLocation As typeMapCoordinate = ftnLocationFromEstimatedPose(tbpEstimatedPose, 0, sngBotRadius) ' tbpEstimatedPose.Location 
                tmcCheckLocation = ftnFirstPointOnLine(tmcCheckLocation, tmcTargetLocation)
                Do
                    tmcCheckLocation = ftnNextPointOnLine(tmcCheckLocation, tmcTargetLocation)
                    ftnClearPathFwd_NoTurnCheck = ftnSafeLocationForMove(tbpEstimatedPose, tmcCheckLocation, sngDistanceToTarget)
                Loop Until ftnLocationsEqual(tmcCheckLocation, tmcTargetLocation) Or boolStop Or Not ftnClearPathFwd_NoTurnCheck
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnClearPathFwd_NoTurnCheck")
        End Try
    End Function



    Function ftnMoveXY_ConditionsMet(ByRef tbpEstimatedPose As typeBotPose, ByRef tmcTargetLocation As typeMapCoordinate) As Boolean
        ' this function returns true if the path between tbpEstimatedPose.Location and tmcTargetLocation is clear of obstructions
        ' and not near a wall
        Try
            ftnMoveXY_ConditionsMet = False
            If boolEnableXY_Motion Then
                Dim sngDistanceToTarget As Single = (ftnDistanceBetweenLocations(tbpEstimatedPose.Location, tmcTargetLocation))
                Dim tbpTargetPose As typeBotPose = tbpEstimatedPose
                tbpTargetPose.Location = tmcTargetLocation
                Dim sngTurnAngle As Short = Abs(ftnTurnAngle(tbpEstimatedPose, tbpTargetPose))
                If (sngDistanceToTarget > sngTwentyInches) And (sngTurnAngle < 45) Or _
                    (sngDistanceToTarget > sngFifteenInches) And (sngTurnAngle < 20) Or _
                    (sngTurnAngle < 10) Then
                    ' If (sngDistanceToTarget > 0) And (sngTurnAngle < 45) Then
                    sngDistanceToTarget = sngDistanceToTarget * 2 ' (Very) Rough approximation - includes path curve
                    Dim sngDistanceFromStart As Single = sngThreeInches
                    Dim tbpCheckPose As typeBotPose
                    Do
                        sngDistanceFromStart += sngTwoInches
                        tbpCheckPose = ftnXY_DR(tbpEstimatedPose, sngDistanceFromStart, tbpTargetPose, False)
                        ftnMoveXY_ConditionsMet = ftnSafeLocationForMove(tbpEstimatedPose, tbpCheckPose.Location, sngDistanceFromStart)
                        subDoEventsIfMainThread()
                    Loop Until ftnLocationsEqual(tbpCheckPose.Location, tmcTargetLocation) Or (sngDistanceFromStart > sngDistanceToTarget) Or boolStop Or Not ftnMoveXY_ConditionsMet
                End If
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnMoveXY_ConditionsMet")
        End Try
    End Function



    Function ftnSafeLocationForMove(ByRef tbpStartPose As typeBotPose, ByRef tmcCheckLocation As typeMapCoordinate, ByRef sngDistanceFromStart As Single) As Boolean
        Try
            ftnSafeLocationForMove = True
            Dim intCellDistanceFromWallSquared As Short = intDistanceFromWallSquaredArray(tmcCheckLocation.X, tmcCheckLocation.Y)
            If (intCellDistanceFromWallSquared < intMaxDistanceFromWallSquared) Then
                If (sngDistanceFromStart >= sngfortyInches) Then
                    ftnSafeLocationForMove = False
                ElseIf (sngDistanceFromStart >= sngSixInches) Then
                    ftnSafeLocationForMove = (intCellDistanceFromWallSquared >= intSafeDistanceFromWallSquared)
                ElseIf (intCellDistanceFromWallSquared < intMinDistanceFromWallSquared) Then  ' 
                    ftnSafeLocationForMove = (intCellDistanceFromWallSquared > intDistanceFromWallSquaredArray(tbpStartPose.Location.X, tbpStartPose.Location.Y))
                    'ElseIf (sngDistanceFromStart > sngMaxTravel) Then
                    '    ftnSafeLocationForMove = False
                End If
            End If
        Catch ex As Exception
            ftnSafeLocationForMove = False
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnSafeLocationForMove")
            frmDebug.subWriteLogs(Color.Black, ":  tbpStartPose= " & tbpStartPose.Location.X & ", " & tbpStartPose.Location.Y & "  at " & tbpStartPose.Orientation)
            frmDebug.subWriteLogs(Color.Black, ":  tmcCheckLocation= " & tmcCheckLocation.X & ", " & tmcCheckLocation.Y)
            frmDebug.subWriteLogs(Color.Black, ":  sngDistanceFromStart= " & sngDistanceFromStart)
        End Try
    End Function

End Module
