﻿Imports System.Globalization
Imports System.IO
Imports System.Math
Imports System.EventArgs
Imports System.Threading
Imports System.Net

Module modScanVectors

    Public Function ftnGetScanVectors(ByRef tbpEstimatedPose As typeBotPose, ByRef boolDrawVectors As Boolean) As Boolean
        ' Returns true if scan vectors received
        Try
            ftnGetScanVectors = False
            If Not boolBotIsMoving Then frmRangeFinder.subGetRangerReadings()
            If boolURGConnected Then ' Get URG04 Depth Readings
                If boolDebugURG Then frmDebug.subWriteLogs(Color.Green, ftnElapsedTime(timeMoveCommandSent) & "   URG Scan vectors received")
                ftnGetScanVectors = ftnGetURGScanVectors()
            ElseIf boolKinectConnected Then ' Get Kinect Depth Readings
                If boolDebugKinect Then frmDebug.subWriteLogs(Color.Green, ftnElapsedTime(timeMoveCommandSent) & "  Kinect Scan vectors received")
                ftnGetScanVectors = ftnGetKinectScanVectors()
            ElseIf boolHeroSelected Then
                ftnGetScanVectors = ftnGetHeroScanVectors(tbpEstimatedPose)
            Else
                frmDebug.subWriteLogs(Color.Green, ftnElapsedTime(timeMoveCommandSent) & "  ERROR: Scan vectors unavailable")
                If Not boolBotIsMoving Then
                    ftnSendCommand("play phrase Beep")
                    subWait(200)
                    ftnSendCommand("play phrase Beep")
                End If
                strStatusDriving = "Scanner Error - Please Reconnect "
                GotABotClass.subRedraw(tbpEstimatedPose)
            End If
            If ftnGetScanVectors And Not boolBotIsMoving And boolDrawVectors Then
                GotABotClass.subRedraw(tbpEstimatedPose)
                subDrawScanVectors(tbpEstimatedPose, False)
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnGetScanVectors - " & ex.ToString)
        End Try
    End Function

     Public Sub subCalculateScanBeamAngles()
        Try
            Dim sngBeamOrientationFactor As Single = sngScanFOV / constMaxScannerBeams
            For intBeamNumber As Short = 0 To constMaxScannerBeams - 1
                trvScanVectors(intBeamNumber).sngAngle = (sngScanFOV / 2 - sngBeamOrientationFactor * intBeamNumber) + intScanOffsetDegrees
            Next intBeamNumber
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in subCalculateScanBeamAngles")
        End Try
    End Sub

    Public Sub subCalculateScanCalCoefficients()
        Try
            Dim sngScannerCalibrateCM As Single = sngScannerCalibrate / 10
            Dim sngScannerCurveAdjusted As Single = sngScannerCurveAdjust / ((sngScanFOV / 2) ^ 1.4)
            For intBeamNumber As Short = 0 To constMaxScannerBeams - 1
                sngScannerCalCoefficients(intBeamNumber) = sngScannerCalibrateCM * (1 + sngScannerCurveAdjusted * Abs(trvScanVectors(intBeamNumber).sngAngle) ^ 1.4)
            Next intBeamNumber
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in subCalculateScanCalCoefficients")
        End Try
    End Sub


End Module
