﻿
'TODO - ADD watchdog, 

' FOV = 240 degrees
' 10 Hz Update Rate
' 1mm resolution
' 0.36 degrees angular resolution
' 682 Beams 

' This module makes use of only 640 beams and a  FOV of 225.22 degrees


'SCIP is a protocol used to communicate with URG. 
'Commands of SCIP protocols are used to interact with the URG. 

'SCIP commands are used to interact with URG sensor
'SCIP commands are classified as "Sensor information command", "Distance acquisition command" , and "Mode setting command".
'Setting of operation mode
'   "SCIP2.0" ... Change SCIP2.0 mode
'   "TM0", "TM1", "TM2" ... Time stamp mode
'   "SS" ... Change the baudrate
'   "BM" ... laser on.
'   "QT" ... laser off
'   "RS" ... Resets parameter, measurement suspends.
'   "CR" ... Change the rotation speed of the motor.
'Obtain status information
'   "VV" ... Acquisition of version information.
'   "PP" ... Acquisition of parameter information.
'   "II" ... Acquisition of status.
'Receive distance data
'   "GD", "GS" ... Acquisition of distance data one at a time.
'   "MD", "MS" ... Acquisition of distance data continuously.

Imports System
Imports System.Math
Imports System.Net
Imports System.Threading
Imports System.IO.Ports


Module modURG
    Friend WithEvents comURG As New SerialPort
    Dim strMessage As String
    Dim listData As New List(Of Integer)

    Public Function ftnURGInitializeTop() As Boolean
        Try
            ftnURGInitializeTop = False
            If Not boolURGConnected Then
                Dim availableSerialPorts As System.Collections.ObjectModel.ReadOnlyCollection(Of String) = My.Computer.Ports.SerialPortNames
                If Not boolShutDown And (strURGComPort <> "None") And availableSerialPorts.Contains(strURGComPort) Then
                    If Not comURG.IsOpen Then
                        With comURG
                            .PortName = strURGComPort
                            '.BaudRate = 115200
                            '.DataBits = 8
                            '.Parity = Parity.None
                            '.StopBits = StopBits.One
                            '.Handshake = Handshake.XOnXOff
                        End With
                        comURG.Open()
                    End If
                    If Not comURG.IsOpen Then
                        ftnSendCommand("play phrase Beep")
                        MsgBox(" ERROR:  URG 04 not detected.")
                        boolURGConnected = False
                        'subClosecomURG()
                    Else
                        sngScanFOV = 225.22
                        intScanVectorsMaxRangeInches = 560 / 2.54
                        intScanVectorsMinRangeInches = 0
                        subCalculateScanBeamAngles()
                        subCalculateScanCalCoefficients()
                        comURG.Write("SCIP2.0" & vbLf & "RS" & vbLf & "QT" & vbLf)

                        Dim ye As String = comURG.ReadExisting().Trim().Replace(vbLf, ";").Replace(";;", ";")

                        While (ye.EndsWith("00P"))
                            comURG.Write("QT" & vbLf)
                            ye = comURG.ReadExisting().Trim().Replace(vbLf, ";").Replace(";;", ";")
                        End While
                        boolURGConnected = ftnGetURGScanVectors()
                    End If
                End If
            End If
            ftnURGInitializeTop = boolURGConnected
            If ftnURGInitializeTop Then intScannerDelay = 500
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnURGInitializeTop " & ex.ToString)
            boolURGConnected = False
            ftnURGInitializeTop = False
        End Try
    End Function




    Private Sub subURGportDataReceived(sender As System.Object, e As System.IO.Ports.SerialDataReceivedEventArgs) Handles comURG.DataReceived
        Try
            strMessage = comURG.ReadExisting()
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in subURGportDataReceived " & ex.ToString)
        End Try
    End Sub

    Public Function ftnGetURGScanVectors() As Boolean
        ftnGetURGScanVectors = False
        Try
            If comURG.IsOpen Then
                strMessage = ""
                comURG.Write("BM" & vbLf) ' Laser on
                subWait(150)
                comURG.Write("GD0044072501" & vbCrLf)
                Dim timeOut As DateTime = Now().AddMilliseconds(1000)
                Do Until ((strMessage.StartsWith("GD") And (strMessage.Length > 2133)) Or boolShutDown Or (Now() > timeOut))
                    System.Windows.Forms.Application.DoEvents()
                    subWait(20)
                Loop
                comURG.Write("QT" & vbLf) ' Laser off
                If (Now() > timeOut) Then
                    ftnGetURGScanVectors = False
                    boolURGConnected = False
                Else
                    Dim strTemp() As String = strMessage.Split(vbLf)
                    Dim strLine As String = ""
                    listData.Clear()
                    For i As Integer = 0 To strTemp.Length - 1
                        If (i > 2 And strTemp(i).Length > 0) Then
                            strLine = strLine & strTemp(i).Substring(0, strTemp(i).Length - 1).ToString()
                        End If
                    Next

                    strLine = strLine.Substring(0, strLine.Length - 1)
                    listData.AddRange(ftnDecodeStr(strLine))
                    For intBeamNumber As Short = 0 To constMaxScannerBeams - 1  ' Convert to trvScanVectors and apply correction coefficients
                        trvScanVectors(intBeamNumber).sngRange = listData(660 - intBeamNumber)
                        trvScanVectors(intBeamNumber).sngRange = ftnCmToMapUnits(trvScanVectors(intBeamNumber).sngRange) * sngScannerCalCoefficients(intBeamNumber)
                    Next intBeamNumber
                    ftnGetURGScanVectors = True
                End If
            End If
        Catch ex As Exception
            comURG.Write("QT" & vbLf) ' Laser off
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnGetURGScanVectors " & ex.ToString)
        End Try
    End Function


    Function ftnDecodeStr(strMessageLine As String) As List(Of Integer)
        Dim listDecode As New List(Of Integer)
        Try
            Dim strTemp As String = ""
            For i As Integer = 0 To strMessageLine.Length Step 3
                If i < strMessageLine.Length - 3 Then
                    strTemp = strMessageLine.Substring(i, 3)
                Else
                    strTemp = strMessageLine.Substring(i)
                End If
                listDecode.Add(ftnDecode(strTemp))
            Next
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnDecodeStr " & ex.ToString)
        End Try
        Return listDecode
    End Function

    Function ftnDecode(strSixBitEncode As String) As Integer
        Dim intDecode As Integer = 0
        Try
            For Each c As Char In strSixBitEncode
                intDecode <<= 6
                intDecode = intDecode And (Not 63)
                intDecode = intDecode Or (Asc(c) - 48)
            Next
        Catch ex As Exception
            frmDebug.subWriteLogs(Color.Red, " EXCEPTION in ftnDecode " & ex.ToString)
        End Try
        Return intDecode
    End Function


  

End Module
